import os
from typing import Dict, List, Literal, Optional, Set, Union, Callable
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import pandas as pd
from slugify import slugify
from loguru import logger

from syngen.train import launch_train
from syngen.infer import launch_infer
from syngen.ml.data_loaders import DataEncryptor, DataLoader, MetadataLoader
from syngen.ml.utils import (
    fetch_config,
    fetch_env_variables,
    get_reports,
    setup_log_process
)
from syngen.ml.reporters import (
    Report,
    AccuracyReporter,
    SampleAccuracyReporter,
)
from syngen.ml.validation_schema import ValidationSchema, ReportTypes
from syngen.ml.context import global_context, get_context


class BaseDataIO(ABC):
    """
    Base class for DataIO and AdvancedDataIO to handle common initialization logic
    """
    def __init__(self, path: str, fernet_key: Optional[str] = None, **kwargs):
        self.path = path
        self.fernet_key = fernet_key
        self.format_settings = kwargs
        self.metadata = self._create_metadata()
        self._validate_metadata()
        global_context(metadata=kwargs)
        self._create_data_loader()

    @abstractmethod
    def _create_metadata(self):
        pass

    @abstractmethod
    def _validate_metadata(self):
        pass

    @abstractmethod
    def _create_data_loader(self):
        pass


class DataIO(BaseDataIO):
    """
    SDK class for loading and saving data with optional encryption and format settings.
    """
    def __init__(self, path: str, fernet_key: Optional[str] = None, **kwargs):
        super().__init__(path, fernet_key, **kwargs)

    def _create_data_loader(self):
        self.data_loader = DataLoader(
            path=self.path,
            table_name="table",
            metadata=self.metadata
        )

    def _create_metadata(self) -> Dict:
        """
        Create the metadata dictionary for data loading and saving
        """
        return {
            "table": {
                "train_settings": {
                    "source": self.path
                },
                "encryption": {
                    "fernet_key": (
                        fetch_env_variables({"fernet_key": self.fernet_key}).get("fernet_key")
                    )
                },
                "format": self.format_settings
            }
        }

    def _validate_metadata(self):
        ValidationSchema(
            metadata=self.metadata,
            validation_of_source=True,
            process="train"
        ).validate_schema()

    def load_data(self, **kwargs) -> pd.DataFrame:
        """
        Load data from the specified path with optional parameters
        """
        df, _ = self.data_loader.load_data(**kwargs)
        return df

    def load_schema(self):
        """
        Load the original schema of the data
        """
        return self.data_loader.original_schema

    def save_data(
        self,
        df: pd.DataFrame,
        **kwargs
    ):
        """
        Save data to the specified path with optional parameters
        """
        self.data_loader.save_data(
            df,
            schema=kwargs.get("schema"),
            format=get_context().get_config()
        )


@dataclass
class Syngen:
    """
    SDK class for training, inference, and generation of reports.

    Attributes:
        list_of_tables (List[str]): The list of table names.
        metadata_path (Optional[str]): The path to the metadata file.
        table_name (Optional[str]): The name of the table.
        source (Optional[str]): The source of the data.
        loader (Optional[Callable[[str], pd.DataFrame]]): The custom data loader function
        that returns a sample of an original data
        execution_artifacts (Dict): The dictionary to store the information about paths
        to execution artifacts.
        report_types (ReportTypes): The object containing the information
        about available report types.
    """
    list_of_tables: List[str] = field(default_factory=list)
    metadata_path: Optional[str] = None
    table_name: Optional[str] = None
    source: Optional[str] = None
    loader: Optional[Callable[[str], pd.DataFrame]] = None
    execution_artifacts: Dict = field(default_factory=dict)
    report_types: object = field(init=False)

    def _validate_required_parameters(self):
        """
        Validate that required parameters are provided
        """
        if not self.metadata_path and not (self.source or self.loader) and not self.table_name:
            raise AttributeError(
                "It seems that the information about 'metadata_path' or 'table_name' "
                "and 'source' (or 'loader') is absent. Please provide either the information "
                "about 'metadata_path' or the information about 'source' (or 'loader') "
                "and 'table_name'."
            )
        elif not self.metadata_path and (self.source or self.loader) and not self.table_name:
            raise AttributeError(
                "It seems that the information about 'metadata_path' or 'table_name' is absent. "
                "Please provide either the information about 'metadata_path' or "
                "the information about 'source' (or 'loader') and 'table_name'."
            )
        elif not self.metadata_path and self.table_name and not (self.source or self.loader):
            raise AttributeError(
                "It seems that the information about 'metadata_path' or 'source' (or 'loader') "
                "is absent. Please provide either the information about 'metadata_path' or "
                "the information about 'source' (or 'loader') and 'table_name'."
            )
        elif self.metadata_path and self.table_name and self.source and self.loader:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'table_name' and 'source' "
                "and 'loader' will be ignored."
            )
        elif self.metadata_path and self.table_name and self.source:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'source' and 'table_name' will be ignored."
            )
        elif self.metadata_path and self.loader and self.source:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'source' and 'loader' will be ignored."
            )
        elif self.metadata_path and self.loader and self.table_name:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'loader' and 'table_name' will be ignored."
            )
        elif self.metadata_path and self.source:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'source' will be ignored."
            )
        elif self.metadata_path and self.table_name:
            logger.warning(
                "The information about 'metadata_path' was provided. "
                "In this case the information about 'table_name' will be ignored."
            )
        elif self.source and self.loader and self.table_name:
            raise AttributeError(
                "The information about both 'source' and 'loader' was provided. "
                "Please provide only one of them along with 'table_name'."
            )

    def __post_init__(self):
        self._validate_required_parameters()
        self.list_of_tables = (
            [
                table_name
                for table_name in MetadataLoader(path=self.metadata_path).load_data().keys()
                if table_name != "global"
            ]
            if self.metadata_path is not None
            else [self.table_name]
        )
        self.report_types = ReportTypes()

    def _set_execution_artifacts(
        self,
        type_of_process: str,
        table_name: Optional[str] = None,
        reports: List[str] = list()
    ):
        self.execution_artifacts = dict()

        if type_of_process == "report" and reports:
            self._set_report_artifacts(table_name, reports)
        else:
            self._set_process_artifacts(type_of_process)

    def _set_report_artifacts(self, table_name: str, reports: List[str]):
        """
        Set execution artifacts for report generation
        """
        self.execution_artifacts[table_name] = {"generated_reports": {}}
        for report in reports:
            if report == "sample":
                reports = self._get_train_reports(table_name)
                self.execution_artifacts[table_name]["generated_reports"]["sample_report"] = (
                    reports.get("sample_report")
                )

            if report in self.report_types.full_list_of_infer_report_types:
                reports = self._get_infer_reports(table_name)
                self.execution_artifacts[table_name]["generated_reports"][f"{report}_report"] = (
                    reports.get(f"{report}_report")
                )

    def _set_process_artifacts(self, type_of_process: str):
        """
        Set execution artifacts for 'train' or 'infer' processes
        """
        for table_name in self.list_of_tables:
            if type_of_process == "train":
                self.execution_artifacts[table_name] = self._get_train_artifacts(table_name)
            elif type_of_process == "infer":
                self.execution_artifacts[table_name] = self._get_infer_artifacts(table_name)

    def _get_train_reports(self, table_name: str) -> Dict:
        """
        Get generated reports generated during a training process for a certain table
        """
        train_config = self._get_config(table_name, type_of_process="train")
        return train_config.paths["generated_reports"]

    @staticmethod
    def _get_config(table_name: str, type_of_process: str):
        """
        Get configuration for a certain table and process type
        """
        if type_of_process == "train":
            path_to_config = (
                f"model_artifacts/resources/{slugify(table_name)}/"
                "vae/checkpoints/train_config.pkl"
            )
        elif type_of_process == "infer":
            path_to_config = (
                f"model_artifacts/tmp_store/{slugify(table_name)}/infer_config.pkl"
            )
        else:
            raise ValueError(f"Unsupported process type: {type_of_process}")

        config = fetch_config(config_pickle_path=path_to_config)
        return config

    def _get_infer_reports(self, table_name: str) -> Dict:
        """
        Get the list of generated reports generated during an inference process for a certain table
        """
        infer_config = self._get_config(table_name, type_of_process="infer")
        type_of_process = infer_config.type_of_process
        if type_of_process == "train":
            train_config = self._get_config(table_name, type_of_process="train")
            return train_config.paths["generated_reports"]
        return infer_config.paths["generated_reports"]

    def _get_train_artifacts(self, table_name: str) -> Dict:
        """
        Get execution artifacts for a training process for a certain table
        """
        train_config = self._get_config(table_name, type_of_process="train")
        return {
            "losses_path": train_config.paths["losses_path"],
            "path_to_input_data": train_config.paths["input_data_path"],
            "generated_reports": train_config.paths["generated_reports"],
        }

    def _get_infer_artifacts(self, table_name: str) -> Dict:
        """
        Get execution artifacts for an inference process for a certain table
        """
        infer_config = self._get_config(table_name, type_of_process="infer")
        return {
            "path_to_input_data": infer_config.paths["input_data_path"],
            "path_to_generated_data": infer_config.paths["path_to_merged_infer"],
            "generated_reports": infer_config.paths["generated_reports"],
        }

    def train(
        self,
        epochs: int = 10,
        drop_null: bool = False,
        row_limit: Optional[int] = None,
        reports: Union[str, List[str]] = "none",
        log_level: Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        batch_size: int = 32,
        fernet_key: Optional[str] = None
    ):
        launch_train(
            metadata_path=self.metadata_path,
            source=self.source,
            table_name=self.table_name,
            epochs=epochs,
            drop_null=drop_null,
            row_limit=row_limit,
            reports=reports,
            log_level=log_level,
            batch_size=batch_size,
            fernet_key=fernet_key,
            loader=self.loader
        )
        self._set_execution_artifacts(type_of_process="train")

    def infer(
        self,
        size: int = 100,
        run_parallel: bool = False,
        batch_size: int = None,
        random_seed: Optional[int] = None,
        reports: Union[str, List[str]] = "none",
        log_level: Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        fernet_key: Optional[str] = None
    ):
        launch_infer(
            metadata_path=self.metadata_path,
            size=size,
            table_name=self.table_name,
            run_parallel=run_parallel,
            batch_size=batch_size,
            reports=reports,
            random_seed=random_seed,
            log_level=log_level,
            fernet_key=fernet_key,
            loader=self.loader
        )
        self._set_execution_artifacts(type_of_process="infer")

    @staticmethod
    def _validate_artifacts(
        table_name: str,
        completed_processes: Set[str]
    ):
        errors: List[str] = []

        slug = slugify(table_name)

        # Type-specific validations
        path_to_train_config = (
            f"model_artifacts/resources/{slug}/vae/checkpoints/train_config.pkl"
        )
        path_to_train_success_file = f"model_artifacts/resources/{slug}/train_message.success"
        path_to_infer_config = f"model_artifacts/tmp_store/{slug}/infer_config.pkl"
        path_to_infer_success_file = f"model_artifacts/tmp_store/{slug}/infer_message.success"
        checks_of_artifacts = {
            "train": [
                (
                    path_to_train_config,
                    (
                        f"The training configuration wasn't found for the table '{table_name}' "
                        f"in the path - {path_to_train_config}."
                    ),
                ),
                (
                    path_to_train_success_file,
                    (
                        f"The training success file wasn't found for the table '{table_name}' "
                        f"in the path - {path_to_train_success_file}."
                    ),
                ),
            ],
            "infer": [
                (
                    path_to_infer_config,
                    (
                        f"The inference configuration wasn't found for the table '{table_name}' "
                        f"in the path - {path_to_infer_config}"
                    ),
                ),
                (
                    path_to_infer_success_file,
                    (
                        f"The inference success file wasn't found for the table '{table_name}' "
                        f"in the path - {path_to_infer_success_file}."
                    ),
                ),
            ],
        }
        for process in completed_processes:
            for path, message in checks_of_artifacts.get(process, []):
                if not os.path.exists(path):
                    errors.append(message)

        if errors:
            raise FileNotFoundError("\n".join(errors))

    def __get_sample_reporter(
        self,
        table_name: str,
        fernet_key: Optional[str]
    ) -> SampleAccuracyReporter:
        """
        Return a SampleAccuracyReporter instance for the specified table
        """
        train_config = self._get_config(table_name, type_of_process="train")
        train_config.metadata[table_name]["encryption"]["fernet_key"] = fernet_key
        return SampleAccuracyReporter(
            table_name=table_name,
            paths=train_config.paths,
            config=train_config.to_dict(),
            metadata=train_config.metadata,
            loader=train_config.loader,
        )

    def __get_accuracy_reporter(
        self,
        table_name: str,
        report: str,
        fernet_key: Optional[str]
    ) -> AccuracyReporter:
        infer_config = self._get_config(table_name, type_of_process="infer")
        infer_config.reports = [report]
        infer_config.metadata[table_name]["encryption"]["fernet_key"] = fernet_key
        return AccuracyReporter(
            table_name=table_name,
            paths=infer_config.paths,
            config=infer_config.to_dict(),
            metadata=infer_config.metadata,
            type_of_process=infer_config.type_of_process,
            loader=infer_config.loader
        )

    def _register_reporter(self, table_name: str, report: str, fernet_key: Optional[str]) -> None:
        """
        Register a specific type of report
        """
        if report in ["accuracy", "metrics_only"]:
            accuracy_reporter = self.__get_accuracy_reporter(table_name, report, fernet_key)
            Report().register_reporter(table=table_name, reporter=accuracy_reporter)
        elif report == "sample":
            sample_reporter = self.__get_sample_reporter(table_name, fernet_key)
            Report().register_reporter(table=table_name, reporter=sample_reporter)

    def generate_quality_reports(
        self,
        table_name: str,
        reports: Union[List[str], str],
        log_level: Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        fernet_key: Optional[str] = None
    ) -> None:
        """
        Generate reports for a table using existing artifacts to
        evaluate the quality of the input data or the generated data

        Parameters
        ----------
        table_name: str
            Table name to generate the report for
            (same as used in the training or in the inference processes).
        reports: Union[List[str], str]
            List of report types to generate.
        fernet_key: Optional[str]
            the name of the environment variable kept the value of the Fernet key
            for decrypting the input of the original data, if applicable.
        """
        setup_log_process(
            type_of_process="report",
            log_level=log_level,
            table_name=table_name,
            metadata_path=None,
        )

        reports = get_reports(reports, self.report_types, "train")

        if reports:
            if fernet_key is not None:
                fernet_key = fetch_env_variables({"fernet_key": fernet_key}).get("fernet_key")
                DataEncryptor.validate_fernet_key(fernet_key)
            self._validate_artifacts(
                table_name=table_name,
                completed_processes={
                    "infer"
                    if report in self.report_types.infer_report_types
                    else "train"
                    for report in reports
                }
            )

        for report in reports:
            self._register_reporter(table_name, report, fernet_key)

        if reports:
            Report().generate_report()
            Report().clear_report()
        else:
            self.report_types.train_report_types.extend(["none", "all"])
            list_of_available_reports = (
                ", ".join(f"'{rt}'" for rt in self.report_types.train_report_types)
            )
            logger.warning(
                "No reports to generate. Please specify the report type "
                f"from {list_of_available_reports}."
            )

        self._set_execution_artifacts(
            type_of_process="report",
            table_name=table_name,
            reports=reports
        )



from syngen.ml.validation_schema.validation_schema import (  # noqa: F401;
    TrainingSettingsSchema,
    InferSettingsSchema,
    CSVFormatSettingsSchema,
    ExcelFormatSettingsSchema,
    GlobalSettingsSchema,
    ConfigurationSchemaWithoutSource,
    ConfigurationSchemaWithOptionalSource,
    ConfigurationSchemaWithSource,
    BaseConfigurationSchema,
    KeysSchema,
    ValidationSchema,
    SUPPORTED_EXCEL_EXTENSIONS,
    ReportTypes
)



from typing import Dict, Optional, List, Tuple, Set, Literal
import pickle
from uuid import UUID
from datetime import datetime
import base32_crockford
from collections import Counter
import re
from dataclasses import dataclass, field
from copy import deepcopy

import numpy as np
import dill
import pandas as pd
from pandas.tseries.api import guess_datetime_format
from scipy.stats import gaussian_kde
import tqdm
from loguru import logger

from syngen.ml.vae.models.features import (
    CategoricalFeature,
    CharBasedTextFeature,
    EmailFeature,
    ContinuousFeature,
    DateFeature,
    BinaryFeature,
)
from syngen.ml.utils import (
    get_nan_labels,
    nan_labels_to_float,
    get_date_columns,
    fetch_timezone,
    fetch_config,
    TIMEZONE_REGEX
)
from syngen.ml.utils import slugify_parameters
from syngen.ml.utils import clean_up_metadata
from syngen.ml.mlflow_tracker import MlflowTracker


@dataclass
class Dataset:
    df: pd.DataFrame
    schema: Dict
    metadata: Dict
    table_name: str
    main_process: str
    paths: Dict[str, str]
    features: Dict = field(default_factory=dict)
    columns: Dict = field(default_factory=dict)
    is_fitted: bool = field(default=False)
    all_columns: List = field(default_factory=list)
    null_num_column_names: List = field(default_factory=list)
    zero_num_column_names: List = field(default_factory=list)
    uuid_columns: Set = field(default_factory=set)
    uuid_columns_types: Dict = field(default_factory=dict)
    tech_columns: Set = field(default_factory=set)
    custom_categorical_columns: Set = field(default_factory=set)
    categorical_columns: Set = field(default_factory=set)
    str_columns: Set = field(default_factory=set)
    float_columns: Set = field(default_factory=set)
    int_columns: Set = field(default_factory=set)
    date_columns: Set = field(default_factory=set)
    date_mapping: Dict = field(default_factory=dict)
    binary_columns: Set = field(default_factory=set)
    email_columns: Set = field(default_factory=set)
    long_text_columns: Set = field(default_factory=set)
    primary_keys_mapping: Dict = field(default_factory=dict)
    primary_keys_list: List = field(default_factory=list)
    primary_key_name: Optional[str] = field(default=None)
    pk_columns: List = field(default_factory=list)
    unique_keys_mapping: Dict = field(default_factory=dict)
    unique_keys_mapping_list: List = field(default_factory=list)
    unique_keys_list: List = field(default_factory=list)
    uq_columns_lists: List = field(default_factory=list)
    uq_columns: List = field(default_factory=list)
    foreign_keys_mapping: Dict = field(default_factory=dict)
    foreign_keys_list: List = field(default_factory=list)
    fk_columns: List = field(default_factory=list)
    keys_mapping: Dict = field(default_factory=dict)
    fields: Dict = field(default_factory=dict)
    schema_format: str = field(default=str())
    order_of_columns: List = field(default_factory=list)
    nan_labels_dict: Dict = field(default_factory=dict)
    nan_labels_in_uuid: Dict = field(default_factory=dict)
    cast_to_integer: Set = field(default_factory=set)
    cast_to_float: Set = field(default_factory=set)
    dropped_columns: Set = field(default_factory=set)
    format: Dict = field(default_factory=dict)
    to_datetime_conversion: Dict = field(default_factory=dict)

    def _select_str_columns(self) -> List[str]:
        """
        Select the text columns
        """
        if self.schema_format == "CSV":
            text_columns = [
                col for col, dtype in dict(self.df.dtypes).items()
                if dtype in ["object", "string", pd.StringDtype()]
            ]
        else:
            text_columns = [
                col
                for col, data_type in self.fields.items()
                if data_type == "string"
            ]
        return text_columns

    def __post_init__(self):
        self.fields = self.schema.get("fields", {})
        self.schema_format = self.schema.get("format")
        self.order_of_columns: List = fetch_config(
            config_pickle_path=self.paths["initial_order_of_columns_path"]
        )
        self.dropped_columns: Set = {
            column for column in self.fields if self.fields[column] == "removed"
        }
        self.format = self.metadata[self.table_name].get("format", {})

    def _detect_categorical_columns(self):
        """
        Define binary and categorical columns
        """
        self._set_binary_columns()
        self._set_categorical_columns()
        self.binary_columns -= self.categorical_columns

    def _preprocess_df(self, excluded_columns: Set[str]):
        """
        Preprocess the dataframe
        """
        self.nan_labels_dict = get_nan_labels(self.df, excluded_columns)
        self.df = nan_labels_to_float(self.df, self.nan_labels_dict)
        self._cast_to_numeric(excluded_columns)

    def _preparation_step(self):
        """
        Define binary and categorical columns,
        preprocess the dataframe before the detection of data types of columns
        """
        table_config = self.metadata.get(self.table_name, {})
        self._set_non_existent_columns(table_config)
        self._update_metadata(table_config)
        self._set_metadata()
        self._detect_categorical_columns()
        excluded_columns = set().union(
            self.categorical_columns,
            self.binary_columns
        )
        self._preprocess_df(excluded_columns)
        self._update_schema()

    def _cast_to_numeric(self, excluded_columns: Set[str]):
        """
        Cast the values in the column to 'integer' or 'float' data type
        in case all of them might be cast to this data type
        """
        text_columns = self._select_str_columns()
        list_of_columns = set(text_columns) - excluded_columns
        for column in list_of_columns:
            try:
                if self.df[column].dropna().apply(lambda x: float(x).is_integer()).all():
                    self.df[column] = pd.to_numeric(self.df[column], downcast="integer")
                    self.cast_to_integer.add(column)
                elif self.df[column].dropna().apply(lambda x: not float(x).is_integer()).any():
                    self.df[column] = pd.to_numeric(self.df[column], downcast="float")
                    self.cast_to_float.add(column)
            except ValueError:
                continue
        if self.cast_to_integer:
            columns = [f"'{item}'" for item in self.cast_to_integer]
            logger.info(
                f"The columns - {', '.join(columns)} "
                "have been cast to the 'integer' data type"
            )
        if self.cast_to_float:
            columns = [f"'{item}'" for item in self.cast_to_float]
            logger.info(
                f"The columns - {', '.join(columns)} "
                "have been cast to the 'float' data type"
            )

    def __getstate__(self) -> Dict:
        """
        Return a dictionary of the dataset's state
        """
        clean_up_metadata(deepcopy(self.metadata))
        dataset_instance = self.__dict__.copy()
        attribute_keys_to_remove = ["df", "non_existent_columns"]

        for attr_key in attribute_keys_to_remove:
            if attr_key in dataset_instance:
                del dataset_instance[attr_key]
        return dataset_instance

    def __set_pk_key(self, config_of_keys: Dict):
        """
        Set up primary key for the table
        """
        self.primary_keys_mapping = {
            key: value
            for (key, value) in config_of_keys.items()
            if config_of_keys.get(key).get("type") == "PK"
        }

        self.primary_keys_list = list(self.primary_keys_mapping.keys())
        self.primary_key_name = self.primary_keys_list[0] if self.primary_keys_list else None
        pk_columns_lists = [val["columns"] for val in self.primary_keys_mapping.values()]
        self.pk_columns = [col for uq_cols in pk_columns_lists for col in uq_cols]

        if self.primary_key_name is None:
            logger.info("No primary key was set.")
        if self.primary_key_name:
            logger.info(f"The primary key name was set: {self.primary_key_name}")
            self._validate_pk_key()

    def _validate_pk_key(self):
        """
        Check null values and uniqueness in primary key
        """
        errors = []
        # Check NA values in primary key columns
        if self.df[self.pk_columns].isna().any(axis=None):
            pk_columns_with_na = [
                column for column in self.pk_columns if self.df[column].isna().any()
            ]
            error_msg = (
                f"The primary key '{self.primary_key_name}' "
                f"contains null values in columns: {pk_columns_with_na}. "
            )
            errors.append(error_msg)

        # Check uniqueness of primary key
        if self.df[self.pk_columns].duplicated().any():
            error_msg = (
                f"The primary key '{self.primary_key_name}' "
                f"contains duplicates. "
            )
            errors.append(error_msg)

        if errors:
            raise ValueError(
                " ".join(errors) + "Please check the original data."
            )
        else:
            logger.info("Values in primary key are unique.")

    def __set_uq_keys(self, config_of_keys: Dict):
        """
        Set up unique keys for the table
        """
        self.unique_keys_mapping = {
            key: value
            for (key, value) in config_of_keys.items()
            if config_of_keys.get(key).get("type") == "UQ"
        }
        self.unique_keys_mapping_list = list(self.unique_keys_mapping.keys())
        self.unique_keys_list = (
            self.unique_keys_mapping_list if self.unique_keys_mapping_list else []
        )
        self.uq_columns_lists = [
            val["columns"] for val in self.unique_keys_mapping.values()
        ]
        self.uq_columns = [
            col for uq_cols in self.uq_columns_lists for col in uq_cols
        ]

        if not self.unique_keys_list:
            logger.info("No unique keys were set.")
        if self.unique_keys_list:
            logger.info(f"The unique keys were set: {self.unique_keys_list}")
            self._validate_uq_keys()

    def _validate_uq_keys(self):
        """
        Check null values and uniqueness in unique keys
        """
        uq_keys_mapping = dict(
            zip(self.unique_keys_mapping_list, self.uq_columns_lists)
        )
        errors = []
        for key_name, key_columns in uq_keys_mapping.items():
            # explicitly check for > 1 null values in unique key columns
            all_na_mask = self.df[key_columns].isna().all(axis=1)
            if all_na_mask.sum() > 1:
                error_msg = (
                    f"The unique key '{key_name}' contains > 1 null values. "
                )
                errors.append(error_msg)

            not_all_na_mask = ~all_na_mask
            if self.df[not_all_na_mask][key_columns].duplicated().any():
                error_msg = (
                    f"Values in the unique key '{key_name}' are not unique. "
                )
                errors.append(error_msg)

        if errors:
            raise ValueError(
                " ".join(errors) + "Please check the original data."
            )
        else:
            logger.info("Values in unique keys are unique.")

    def _filter_dropped_keys(self, config_of_keys: Dict, type_of_key: str) -> Tuple[Dict, Set]:
        """
        Filter out keys that contain empty columns
        """
        filtered_keys = {}
        dropped_keys = set()

        for key, value in config_of_keys.items():
            if value.get("type") == type_of_key:
                if any(
                    column for column in value.get("columns") if column in self.dropped_columns
                ):
                    dropped_keys.add(key)
                else:
                    filtered_keys[key] = value

        return filtered_keys, dropped_keys

    def __set_fk_keys(self, config_of_keys: Dict):
        """
        Set up foreign keys for the table
        """
        self.foreign_keys_mapping, dropped_fk_keys = self._filter_dropped_keys(
            config_of_keys, "FK"
        )
        self.foreign_keys_list = list(self.foreign_keys_mapping.keys())
        fk_columns_lists = [val["columns"] for val in self.foreign_keys_mapping.values()]
        self.fk_columns = [col for fk_cols in fk_columns_lists for col in fk_cols]

        if dropped_fk_keys:
            logger.info(
                f"The following foreign keys were dropped: {', '.join(dropped_fk_keys)} "
                f"as they contain empty columns: "
                f"{', '.join(self.dropped_columns.union(self.fk_columns))}"
            )

        if self.foreign_keys_list:
            logger.info(
                f"The following foreign keys were set: {', '.join(self.foreign_keys_list)}"
            )
        if not self.foreign_keys_list:
            logger.info("No foreign keys were set.")

    def __set_types(self, pk_uq_keys_mapping):
        """
        Set up list of data types of primary and unique keys
        """
        self.pk_uq_keys_types = {}
        for key_name, config in pk_uq_keys_mapping.items():
            key_columns = config.get("columns")
            for column in key_columns:
                column_type = (
                    str
                    if column
                    in (
                        self.str_columns
                        | self.email_columns
                        | self.categorical_columns
                        | self.date_columns
                        | self.long_text_columns
                        | self.uuid_columns
                    )
                    else float
                )
                self.pk_uq_keys_types[column] = column_type

    def __map_text_pk(self):
        for pk, pk_type in self.pk_uq_keys_types.items():
            if pk_type is str:
                mapper = {k: n for n, k in enumerate(self.df[pk])}
                with open(f"{self.paths['fk_kde_path']}{pk}_mapper.pkl", "wb") as file:
                    pickle.dump(mapper, file)

    def _set_metadata(self):
        config_of_keys = self.metadata.get(self.table_name, {}).get("keys")

        if config_of_keys is not None:
            self.__set_pk_key(config_of_keys)
            self.__set_uq_keys(config_of_keys)
            self.__set_fk_keys(config_of_keys)

        self.keys_mapping = {
            "primary key": self.pk_columns,
            "unique key": self.uq_columns,
            "foreign key": self.fk_columns
        }

    def launch_detection(self):
        self._preparation_step()
        self._launch_detection()

    def _launch_detection(self):
        self._common_detection()
        self._detection_pipeline()

    def _common_detection(self):
        """
        Identify and classify data types within the dataset, including
        binary columns, categorical columns, UUID columns, long text columns,
        and email columns.

        This process is agnostic to the file format of the dataset.
        """
        self._set_uuid_columns()
        self._set_long_text_columns()
        self._set_email_columns()
        if self.nan_labels_dict and self.format.get("na_values", []):
            logger.info(
                f"Despite the fact that data loading utilized the 'format' section "
                f"for handling NA values, some values have been detected by the algorithm "
                f"as NA labels in the columns - {self.nan_labels_dict}"
            )

    def _update_schema(self):
        """
        Synchronize the schema of the table with dataframe
        """
        self.fields = {
            column: data_type
            for column, data_type in self.fields.items()
            if column in self.df.columns
        }
        int_fields = {
            column: "int"
            for column, dtype in self.fields.items()
            if column in self.cast_to_integer
        }
        self.fields.update(int_fields)
        float_fields = {
            column: "float"
            for column, dtype in self.fields.items()
            if column in self.cast_to_float
        }
        self.fields.update(float_fields)
        for column in self.nan_labels_dict.keys():
            self.fields[column] = (
                "int" if all(x.is_integer() for x in self.df[column].dropna()) else "float"
            )

    def _check_if_column_not_removed(self, column_type: str):
        """
        Exclude the column from the list of columns
        if it was removed previously as empty column
        """
        columns = getattr(self, f"{column_type}_columns")
        for col in list(columns):
            if col in self.dropped_columns:
                columns.remove(col)
                logger.warning(
                    f"The column '{col}' was excluded from the list of {column_type} columns "
                    f"as far as this column is empty and was removed from the table - "
                    f"'{self.table_name}'"
                )
            continue
        setattr(self, f"{column_type}_columns", columns)

    def _check_if_column_existed(self, column_type: str):
        """
        Exclude the column from the list of columns
        if it doesn't exist in the table
        """
        removed_columns = []
        columns = getattr(self, f"{column_type}_columns")
        for col in columns:
            if col not in self.df.columns:
                removed_columns.append(col)
            continue

        setattr(
            self,
            f"{column_type}_columns",
            set([i for i in columns if i not in removed_columns])
        )

        if removed_columns:
            logger.warning(
                f"The columns - {', '.join(removed_columns)} were mentioned as {column_type} "
                f"in the metadata of the table - '{self.table_name}'. "
                f"It seems that the columns are absent in the table - '{self.table_name}'. "
                f"Please, check the metadata file"
            )

    def _check_if_column_not_key(self, column_type: str, column_list: List, key_type: str):
        """
        Exclude the column from the list of columns
        if it relates to certain type of key
        """
        columns = getattr(self, f"{column_type}_columns")
        columns_to_remove = {column for column in columns if column in column_list}
        for column in columns_to_remove:
            logger.warning(
                f"The column '{column}' was excluded from the list of {column_type} columns "
                f"as far as this column was set as the {key_type} of the table - "
                f"'{self.table_name}'"
            )
            columns.discard(column)
        setattr(
            self,
            f"{column_type}_columns",
            columns
        )

    def _fetch_categorical_columns(self):
        """
        Fetch the categorical columns from the metadata
        """
        metadata_of_table = self.metadata.get(self.table_name)

        if metadata_of_table is not None:
            self.custom_categorical_columns = set(
                metadata_of_table.get("train_settings", {})
                .get("column_types", {})
                .get("categorical", [])
            )
        if self.custom_categorical_columns:
            logger.info(
                f"The columns - {', '.join(self.custom_categorical_columns)} were defined "
                f"as categorical due to the information from the metadata of the table - "
                f"'{self.table_name}'"
            )

    def _validate_columns(self, column_type: str):
        """
        Validate categorical columns
        """
        self._check_if_column_not_removed(column_type)
        self._check_if_column_existed(column_type)
        for key_type, key_list in self.keys_mapping.items():
            self._check_if_column_not_key(
                column_type=column_type,
                column_list=key_list,
                key_type=key_type
            )

    def _set_binary_columns(self):
        """
        Set up the list of binary columns based on the count of unique values in the column
        """
        self.binary_columns = set(
            [col for col in self.df.columns if self.df[col].fillna("?").nunique() == 2]
        )

    def _define_categorical_columns(self):
        """
        Define the list of categorical columns based on the count of unique values in the column
        """
        self.categorical_columns = set(
            [
                col
                for col in self.df.columns
                if self.df[col].fillna("?").nunique() <= 50
                and col not in self.binary_columns
            ]
        )
        self.categorical_columns.update(self.custom_categorical_columns)

    def _set_categorical_columns(self):
        """
        Set up the list of categorical columns
        """
        self._fetch_categorical_columns()
        self._define_categorical_columns()
        if self.categorical_columns:
            self._validate_columns(column_type="categorical")

    def _set_long_text_columns(self):
        """
        Set up the list of columns with long texts (> 200 symbols)
        """
        text_columns = self._select_str_columns()
        data_subset = self.df[text_columns]

        if not data_subset.empty:
            data_subset = data_subset.loc[
                :, data_subset.apply(lambda x: (x.astype(str).str.len() > 200).any())
            ]
            self.long_text_columns = set(data_subset.columns)
            self.long_text_columns = (self.long_text_columns - self.categorical_columns
                                      - self.binary_columns)
            if self.long_text_columns:
                list_of_long_text_columns = [f"'{column}'" for column in self.long_text_columns]
                logger.info(
                    f"Please note that the columns - {', '.join(list_of_long_text_columns)} "
                    f"contain long texts (> 200 symbols). Such texts' handling consumes "
                    f"significant resources and results in poor quality content, "
                    f"therefore this column(-s) will be generated using "
                    f"a simplified statistical approach"
                )

    def _set_email_columns(self):
        """
        Set up the list of columns with emails (defined by count of @ symbols)
        """
        text_columns = self._select_str_columns()
        data_subset = self.df[text_columns]

        if not data_subset.empty:
            # @ presents in more than 4/5 of not None values of every column
            # Email pattern
            email_pattern = r'^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}$'

            # Vectorized operation to check for email pattern
            email_mask = data_subset.apply(
                lambda col: col.apply(
                    lambda x: isinstance(x, str) and bool(re.search(email_pattern, x))
                )
            )

            # Count the number of emails in each column
            count_emails = email_mask.sum(axis=0)

            # Adjusted count to inverse 4/5 rule
            adjusted_count = count_emails * 1.25

            # Count non-NA values in each column
            non_na_values_count = data_subset.notna().sum(axis=0)

            # Filter columns where adjusted count is greater than non-NA count
            filter_mask = adjusted_count > non_na_values_count

            # Select columns that match the filter
            email_columns = data_subset.columns[filter_mask]

            # Update the email_columns attribute
            self.email_columns = (
                set(email_columns) - self.categorical_columns - self.binary_columns
            )

    @staticmethod
    def _is_valid_ulid(uuid):
        """
        Check if uuid_to_test is a valid ULID (https://github.com/ulid/spec)
        """
        # ULID pattern check using regex
        try:
            if not re.match(r'^[0123456789ABCDEFGHJKMNPQRSTVWXYZ]{26}$', uuid):
                return
            ulid_timestamp = uuid[:10]
            ulid_timestamp_int = base32_crockford.decode(ulid_timestamp)
            datetime.fromtimestamp(ulid_timestamp_int / 1000.0)
            return "ulid"
        except Exception:
            return

    def _get_uuid_version(self, value: str) -> int | None:
        """Get UUID version if value is a valid UUID"""
        try:
            return UUID(value).version
        except (ValueError, AttributeError, TypeError):
            return None

    def _process_values(self, column: pd.DataFrame):
        """
        Process a text column and collect uuid types in the list - 'result',
        and non uuid values into the set - 'non_uuid_values'
        """
        result = []
        non_uuid_values = set()
        for i in column.dropna().unique():
            uuid_version = self._get_uuid_version(i)
            if uuid_version:
                result.append(uuid_version)
            else:
                if ulid := self._is_valid_ulid(i):
                    result.append(ulid)
                else:
                    non_uuid_values.add(i)
        return result, non_uuid_values

    def _is_valid_uuid(self, column: pd.DataFrame):
        """
        Check if a column is a valid UUID column.
        If there are no NaNs and a single non UUID/ULID value,
        it is treated as a nan label and set as NaN during a training process.
        """
        result, non_uuid_values = self._process_values(column)

        if result:
            return self._handle_non_uuid_values(
                column,
                result,
                non_uuid_values,
            )
        else:
            return

    def _log_warning(self, column, non_uuid_values, contain_nan):
        """
        Build the log warning message related to columns containing UUID/ULID values
        """
        warning_msg = f"Column '{column.name}' contains UUID/ULID values"
        if non_uuid_values:
            non_uuid_values_quoted = [f"'{value}'" for value in non_uuid_values]
            if len(non_uuid_values_quoted) > 10:
                warning_msg += (
                    f", and more than {len(non_uuid_values_quoted)} non-UUID/ULID values"
                )
            else:
                warning_msg += f", and non-UUID/ULID values - {', '.join(non_uuid_values_quoted)}"
        if contain_nan:
            warning_msg += ", and null value/s"
        warning_msg += ". The column will be treated as a text column."
        logger.warning(warning_msg)

    def _handle_non_uuid_values(self, column, result, non_uuid_values):
        """
        Handle a UUID column containing non UUID/ULID values
        """
        contain_nan = column.isnull().sum() > 0
        most_common_uuid_type = max(set(result), key=result.count)
        if not non_uuid_values:
            return most_common_uuid_type

        if not contain_nan and len(non_uuid_values) == 1:
            self.__handle_nan_label_in_uuid(column, non_uuid_values)
            return most_common_uuid_type

        if len(non_uuid_values) > 1 or (contain_nan and non_uuid_values):
            self._log_warning(column, non_uuid_values, contain_nan)
            return

    def __handle_nan_label_in_uuid(self, column, non_uuid_values):
        """
        Replaces the unique non-UUID/ULID value with NaNs
        Updates the nan_labels_in_uuid dictionary and
        adds it to 'nan_labels_dict' dict
        """
        unique_non_uuid = next(iter(non_uuid_values))

        logger.info(
            f"Column '{column.name}' contains a unique non-UUID/ULID "
            f"value '{unique_non_uuid}'. It will be treated "
            f"as a null label and replaced with nulls during the training process."
        )
        self.nan_labels_in_uuid[column.name] = unique_non_uuid
        self.df[column.name].replace(unique_non_uuid, np.nan, inplace=True)
        # update the nan_labels_dict with nan_labels_in_uuid
        self.nan_labels_dict.update(self.nan_labels_in_uuid)

    def _set_uuid_columns(self):
        """
        Set up the list of columns with UUIDs
        """
        text_columns = self._select_str_columns()
        data_subset = self.df[text_columns]

        if not data_subset.empty:
            data_subset = data_subset.apply(self._is_valid_uuid)

            self.uuid_columns_types = dict(data_subset[data_subset.isin([1, 2, 3, 4, 5, "ulid"])])
            self.uuid_columns = set(self.uuid_columns_types.keys())

    def _set_date_columns(self):
        """
        Set up the list of date columns
        """
        self.date_columns = (
            get_date_columns(self.df, list(self.str_columns))
            - self.email_columns
            - self.categorical_columns
            - self.binary_columns
            - self.long_text_columns
        )

    def _remove_non_existent_columns(self, columns: list, key: str, key_type: str) -> list:
        """
        Remove the columns from the table metadata which are absent in the table
        """
        updated_columns = []
        for column in columns:
            if column in self.non_existent_columns:
                logger.warning(
                    f"The column - '{column}' was excluded from the {key_type} - "
                    f"'{key}' as far as this column doesn't exist in the table - "
                    f"'{self.table_name}'"
                )
            else:
                updated_columns.append(column)
        return updated_columns

    def _update_metadata(self, table_config: Dict):
        """
        Update the metadata of the table by removing the columns which are absent in the table
        but mentioned in the metadata
        """
        table_metadata = table_config.get("keys", {})
        for key in list(table_metadata.keys()):
            key_type = table_metadata[key].get("type")
            updated_columns = self._remove_non_existent_columns(
                table_metadata[key].get("columns", []), key, key_type
            )
            table_metadata[key]["columns"] = updated_columns
            if not table_metadata[key]["columns"]:
                logger.warning(
                    f"All columns in the key {key} are empty, "
                    f"so it will be removed from the table's metadata"
                )
                table_metadata[key].pop(key, None)
                self.metadata[self.table_name]["keys"].pop(key, None)

    def _set_non_existent_columns(self, table_config: Dict):
        """
        Set up the list of columns which are absent in the table
        """
        non_existent_columns = {
            column
            for key_config in table_config.get("keys", {}).values()
            for column in key_config.get("columns", [])
            if column not in self.df.columns
        }

        self.non_existent_columns = non_existent_columns - self.dropped_columns

    def __define_date_format(self, column: str) -> str:
        """
        Define the most common date format in the column
        Supported formats:
        - %d-%m-%Y, %d/%m/%Y, %d.%m.%Y, %m-%d-%Y, %m/%d/%Y, %m.%d.%Y, %Y-%m-%d, %Y/%m/%d, %Y.%m.%d;
        - %d %B %Y, %d %b %Y, %Y %B %d, %Y %b %d; %B %d, %Y; %b %d, %Y;
        - %Y-%m-%dT%H:%M:%S, %Y-%m-%d %H:%M:%S, %Y-%m-%d-%H:%M:%S;
        - %Y/%m/%dT%H:%M:%S, %Y/%m/%d %H:%M:%S, %Y/%m/%d-%H:%M:%S;
        - %Y.%m.%dT%H:%M:%S, %Y.%m.%d %H:%M:%S, %Y.%m.%d-%H:%M:%S;
        - %Y%m%dT%H:%M:%S, %Y%m%d %H:%M:%S, %Y%m%d-%H:%M:%S;
        - %d-%m-%YT%H:%M:%S, %d-%m-%Y %H:%M:%S, %d-%m-%Y-%H:%M:%S;
        - %d/%m/%YT%H:%M:%S, %d/%m/%Y %H:%M:%S, %d/%m/%Y-%H:%M:%S;
        - %d.%m.%YT%H:%M:%S, %d.%m.%Y %H:%M:%S, %d.%m.%Y-%H:%M:%S;
        - %m-%d-%YT%H:%M:%S, %m-%d-%Y %H:%M:%S, %m-%d-%Y-%H:%M:%S;
        - %m/%d/%YT%H:%M:%S, %m/%d/%Y %H:%M:%S, %m/%d/%Y-%H:%M:%S;
        - %m.%d.%YT%H:%M:%S, %m.%d.%Y %H:%M:%S, %m.%d.%Y-%H:%M:%S;
        - %Y-%m-%dT%H:%M:%S.%f, %Y-%m-%d %H:%M:%S.%f, %Y-%m-%d-%H:%M:%S.%f;
        - %Y/%m/%dT%H:%M:%S.%f, %Y/%m/%d %H:%M:%S.%f, %Y/%m/%d-%H:%M:%S.%f;
        - %Y.%m.%dT%H:%M:%S.%f, %Y.%m.%d %H:%M:%S.%f, %Y.%m.%d-%H:%M:%S.%f;
        - %Y%m%dT%H:%M:%S.%f, %Y%m%d %H:%M:%S.%f, %Y%m%d-%H:%M:%S.%f;
        - %d-%m-%YT%H:%M:%S.%f, %d-%m-%Y %H:%M:%S.%f, %d-%m-%Y-%H:%M:%S.%f;
        - %d/%m/%YT%H:%M:%S.%f, %d/%m/%Y %H:%M:%S.%f, %d/%m/%Y-%H:%M:%S.%f;
        - %d.%m.%YT%H:%M:%S.%f, %d.%m.%Y %H:%M:%S.%f, %d.%m.%Y-%H:%M:%S.%f;
        - %m-%d-%YT%H:%M:%S.%f, %m-%d-%Y %H:%M:%S.%f, %m-%d-%Y-%H:%M:%S.%f;
        - %m/%d/%YT%H:%M:%S.%f, %m/%d/%Y %H:%M:%S.%f, %m/%d/%Y-%H:%M:%S.%f;
        - %m.%d.%YT%H:%M:%S.%f, %m.%d.%Y %H:%M:%S.%f, %m.%d.%Y-%H:%M:%S.%f;
        - %Y-%m-%dT%H:%M:%S%z, %Y-%m-%d %H:%M:%S%z, %Y-%m-%d-%H:%S%z;
        - %Y/%m/%dT%H:%M:%S%z, %Y/%m/%d %H:%M:%S%z, %Y/%m/%d-%H:%M:%S%z;
        - %Y.%m.%dT%H:%M:%S%z, %Y.%m.%d %H:%M:%S%z, %Y.%m.%d-%H:%M:%S%z;
        - %Y%m%dT%H:%M:%S%z, %Y%m%d %H:%M:%S%z, %Y%m%d-%H:%M:%S%z;
        - %d-%m-%YT%H:%M:%S%z, %d-%m-%Y %H:%M:%S%z, %d-%m-%Y-%H:%M:%S%z;
        - %d/%m/%YT%H:%M:%S%z, %d/%m/%Y %H:%M:%S%z, %d/%m/%Y-%H:%M:%S%z;
        - %d.%m.%YT%H:%M:%S%z, %d.%m.%Y %H:%M:%S%z, %d.%m.%Y-%H:%M:%S%z;
        - %m-%d-%YT%H:%M:%S%z, %m-%d-%Y %H:%M:%S%z, %m-%d-%Y-%H:%M:%S%z;
        - %m/%d/%YT%H:%M:%S%z, %m/%d/%Y %H:%M:%S%z, %m/%d/%Y-%H:%M:%S%z;
        - %m.%d.%YT%H:%M:%S%z, %m.%d.%Y %H:%M:%S%z, %m.%d.%Y-%H:%M:%S%z;
        - %Y-%m-%dT%H:%M:%S %z, %Y-%m-%d %H:%M:%S %z, %Y-%m-%d-%H:%M:%S %z;
        - %Y/%m/%dT%H:%M:%S %z, %Y/%m/%d %H:%M:%S %z, %Y/%m/%d-%H:%M:%S %z;
        - %Y.%m.%dT%H:%M:%S %z, %Y.%m.%d %H:%M:%S %z, %Y.%m.%d-%H:%M:%S %z;
        - %Y%m%dT%H:%M:%S %z, %Y%m%d %H:%M:%S %z, %Y%m%d-%H:%M:%S %z;
        - %d-%m-%YT%H:%M:%S %z, %d-%m-%Y %H:%M:%S %z, %d-%m-%Y-%H:%M:%S %z;
        - %d/%m/%YT%H:%M:%S %z, %d/%m/%Y %H:%M:%S %z, %d/%m/%Y-%H:%M:%S %z;
        - %d.%m.%YT%H:%M:%S %z, %d.%m.%Y %H:%M:%S %z, %d.%m.%Y-%H:%M:%S %z;
        - %m-%d-%YT%H:%M:%S %z, %m-%d-%Y %H:%M:%S %z, %m-%d-%Y-%H:%M:%S %z;
        - %m/%d/%YT%H:%M:%S %z, %m/%d/%Y %H:%M:%S %z, %m/%d/%Y-%H:%M:%S %z;
        - %m.%d.%YT%H:%M:%S %z, %m.%d.%Y %H:%M:%S %z, %m.%d.%Y-%H:%M:%S %z;
        - %Y-%m-%dT%H:%M:%S.%f, %Y-%m-%d %H:%M:%S.%f, %Y-%m-%d-%H:%M:%S.%f;
        - %Y/%m/%dT%H:%M:%S.%f, %Y/%m/%d %H:%M:%S.%f, %Y/%m/%d-%H:%M:%S.%f;
        - %Y.%m.%dT%H:%M:%S.%f, %Y.%m.%d %H:%M:%S.%f, %Y.%m.%d-%H:%M:%S.%f;
        - %Y%m%dT%H:%M:%S.%f, %Y%m%d %H:%M:%S.%f, %Y%m%d-%H:%M:%S.%f;
        - %d-%m-%YT%H:%M:%S.%f, %d-%m-%Y %H:%M:%S.%f, %d-%m-%Y-%H:%M:%S.%f;
        - %d/%m/%YT%H:%M:%S.%f, %d/%m/%Y %H:%M:%S.%f, %d/%m/%Y-%H:%M:%S.%f;
        - %d.%m.%YT%H:%M:%S.%f, %d.%m.%Y %H:%M:%S.%f, %d.%m.%Y-%H:%M:%S.%f;
        - %m-%d-%YT%H:%M:%S.%f, %m-%d-%Y %H:%M:%S.%f, %m-%d-%Y-%H:%M:%S.%f;
        - %m/%d/%YT%H:%M:%S.%f, %m/%d/%Y %H:%M:%S.%f, %m/%d/%Y-%H:%M:%S.%f;
        - %m.%d.%YT%H:%M:%S.%f, %m.%d.%Y %H:%M:%S.%f, %m.%d.%Y-%H:%M:%S.%f;
        - %Y-%m-%dT%H:%M:%S.%f %z, %Y-%m-%d %H:%M:%S.%f %z, %Y-%m-%d-%H:%M:%S.%f %z;
        - %Y/%m/%dT%H:%M:%S.%f %z, %Y/%m/%d %H:%M:%S.%f %z, %Y/%m/%d-%H:%M:%S.%f %z;
        - %Y.%m.%dT%H:%M:%S.%f %z, %Y.%m.%d %H:%M:%S.%f %z, %Y.%m.%d-%H:%M:%S.%f %z;
        - %Y%m%dT%H:%M:%S.%f %z, %Y%m%d %H:%M:%S.%f %z, %Y%m%d-%H:%M:%S.%f %z;
        - %d-%m-%YT%H:%M:%S.%f %z, %d-%m-%Y %H:%M:%S.%f %z, %d-%m-%Y-%H:%M:%S.%f %z;
        - %d/%m/%YT%H:%M:%S.%f %z, %d/%m/%Y %H:%M:%S.%f %z, %d/%m/%Y-%H:%M:%S.%f %z;
        - %d.%m.%YT%H:%M:%S.%f %z, %d.%m.%Y %H:%M:%S.%f %z, %d.%m.%Y-%H:%M:%S.%f %z;
        - %m-%d-%YT%H:%M:%S.%f %z, %m-%d-%Y %H:%M:%S.%f %z, %m-%d-%Y-%H:%M:%S.%f %z;
        - %m/%d/%YT%H:%M:%S.%f %z, %m/%d/%Y %H:%M:%S.%f %z, %m/%d/%Y-%H:%M:%S.%f %z;
        - %m.%d.%YT%H:%M:%S.%f %z, %m.%d.%Y %H:%M:%S.%f %z, %m.%d.%Y-%H:%M:%S.%f %z;
        - %Y-%m-%dT%H:%M:%S.%f%z, %Y-%m-%d %H:%M:%S%f%z, %Y-%m-%d-%H:%M:%S%f%z;
        - %Y/%m/%dT%H:%M:%S.%f%z, %Y/%m/%d %H:%M:%S%f%z, %Y/%m/%d-%H:%M:%S%f%z;
        - %Y.%m.%dT%H:%M:%S.%f%z, %Y.%m.%d %H:%M:%S%f%z, %Y.%m.%d-%H:%M:%S%f%z;
        - %Y%m%dT%H:%M:%S.%f%z, %Y%m%d %H:%M:%S%f%z, %Y%m%d-%H:%M:%S%f%z;
        - %d-%m-%YT%H:%M:%S.%f%z, %d-%m-%Y %H:%M:%S%f%z, %d-%m-%Y-%H:%M:%S%f%z;
        - %d/%m/%YT%H:%M:%S.%f%z, %d/%m/%Y %H:%M:%S%f%z, %d/%m/%Y-%H:%M:%S%f%z;
        - %d.%m.%YT%H:%M:%S.%f%z, %d.%m.%Y %H:%M:%S%f%z, %d.%m.%Y-%H:%M:%S%f%z;
        - %m-%d-%YT%H:%M:%S.%f%z, %m-%d-%Y %H:%M:%S%f%z, %m-%d-%Y-%H:%M:%S%f%z;
        - %m/%d/%YT%H:%M:%S.%f%z, %m/%d/%Y %H:%M:%S%f%z, %m/%d/%Y-%H:%M:%S%f%z;
        - %m.%d.%YT%H:%M:%S.%f%z, %m.%d.%Y %H:%M:%S%f%z, %m.%d.%Y-%H:%M:%S%f%z;
        - "%a, %d %b %Y %H:%M:%S %z", "%a, %d %B %Y %H:%M:%S %z";
        - "%A, %d %b %Y %H:%M:%S %z", "%A, %d %B %Y %H:%M:%S %z";
        - "%a, %d %b %Y %H:%M:%S %z", "%a, %d %B %Y %H:%M:%S.%f %z";
        - "%A, %d %b %Y %H:%M:%S %z", "%A, %d %B %Y %H:%M:%S.%f %z";
        - "%a, %d %b %Y %H:%M:%S %z", "%a, %d %B %Y %H:%M:%S.%f%z";
        - "%A, %d %b %Y %H:%M:%S %z", "%A, %d %B %Y %H:%M:%S.%f%z";
        - "%a, %d %b %Y %H:%M:%S %z", "%a, %d %B %Y %H:%M:%S%z";
        - "%A, %d %b %Y %H:%M:%S %z", "%A, %d %B %Y %H:%M:%S%z";

        Not supported formats (for them the default format will be used):
        - %m-%d-%y, %m/%d/%y, %m.%d.%y, %d-%m-%y, %d/%m/%y, %d.%m.%y, %y-%m-%d, %y/%m/%d, %y.%m.%d;
        - %m-%d-%yT%H:%M:%S, %m-%d-%y %H:%M:%S, %m-%d-%y-%H:%M:%S;
        - %d-%m-%yT%H:%M:%S, %d-%m-%y %H:%M:%S, %d-%m-%y-%H:%M:%S;
        - %y-%m-%dT%H:%M:%S, %y-%m-%d %H:%M:%S, %y-%m-%d-%H:%M:%S;
        - %m/%d/%yT%H:%M:%S, %m/%d/%y %H:%M:%S, %m/%d/%y-%H:%M:%S;
        - %d/%m/%yT%H:%M:%S, %d/%m/%y %H:%M:%S, %d/%m/%y-%H:%M:%S;
        - %y/%m/%dT%H:%M:%S, %y/%m/%d %H:%M:%S, %y/%m/%d-%H:%M:%S;
        - %m.%d.%yT%H:%M:%S, %m.%d.%y %H:%M:%S, %m.%d.%y-%H:%M:%S;
        - %d.%m.%yT%H:%M:%S, %d.%m.%y %H:%M:%S, %d.%m.%y-%H:%M:%S;
        - %y.%m.%dT%H:%M:%S, %y.%m.%d %H:%M:%S, %y.%m.%d-%H:%M:%S;
        - %m-%d-%yT%H:%M:%S.%f, %m-%d-%y %H:%M:%S.%f, %m-%d-%y-%H:%M:%S.%f;
        - %d-%m-%yT%H:%M:%S.%f, %d-%m-%y %H:%M:%S.%f, %d-%m-%y-%H:%M:%S.%f;
        - %y-%m-%dT%H:%M:%S.%f, %y-%m-%d %H:%M:%S.%f, %y-%m-%d-%H:%M:%S.%f;
        - %m/%d/%yT%H:%M:%S.%f, %m/%d/%y %H:%M:%S.%f, %m/%d/%y-%H:%M:%S.%f;
        - %d/%m/%yT%H:%M:%S.%f, %d/%m/%y %H:%M:%S.%f, %d/%m/%y-%H:%M:%S.%f;
        - %y/%m/%dT%H:%M:%S.%f, %y/%m/%d %H:%M:%S.%f, %y/%m/%d-%H:%M:%S.%f;
        - %m.%d.%yT%H:%M:%S.%f, %m.%d.%y %H:%M:%S.%f, %m.%d.%y-%H:%M:%S.%f;
        - %d.%m.%yT%H:%M:%S.%f, %d.%m.%y %H:%M:%S.%f, %d.%m.%y-%H:%M:%S.%f;
        - %y.%m.%dT%H:%M:%S.%f, %y.%m.%d %H:%M:%S.%f, %y.%m.%d-%H:%M:%S.%f;
        - %m-%d-%yT%H:%M:%S.%f %z, %m-%d-%y %H:%M:%S.%f %z, %m-%d-%y-%H:%M:%S.%f %z;
        - %d-%m-%yT%H:%M:%S.%f %z, %d-%m-%y %H:%M:%S.%f %z, %d-%m-%y-%H:%M:%S.%f %z;
        - %y-%m-%dT%H:%M:%S.%f %z, %y-%m-%d %H:%M:%S.%f %z, %y-%m-%d-%H:%M:%S.%f %z;
        - %m/%d/%yT%H:%M:%S.%f %z, %m/%d/%y %H:%M:%S.%f %z, %m/%d/%y-%H:%M:%S.%f %z;
        - %d/%m/%yT%H:%M:%S.%f %z, %d/%m/%y %H:%M:%S.%f %z, %d/%m/%y-%H:%M:%S.%f %z;
        - %y/%m/%dT%H:%M:%S.%f %z, %y/%m/%d %H:%M:%S.%f %z, %y/%m/%d-%H:%M:%S.%f %z;
        - %m.%d.%yT%H:%M:%S.%f %z, %m.%d.%y %H:%M:%S.%f %z, %m.%d.%y-%H:%M:%S.%f %z;
        - %d.%m.%yT%H:%M:%S.%f %z, %d.%m.%y %H:%M:%S.%f %z, %d.%m.%y-%H:%M:%S.%f %z;
        - %y.%m.%dT%H:%M:%S.%f %z, %y.%m.%d %H:%M:%S.%f %z, %y.%m.%d-%H:%M:%S.%f %z;
        - %m-%d-%yT%H:%M:%S.%f%z, %m-%d-%y %H:%M:%S.%f%z, %m-%d-%y-%H:%M:%S.%f%z;
        - %d-%m-%yT%H:%M:%S.%f%z, %d-%m-%y %H:%M:%S.%f%z, %d-%m-%y-%H:%M:%S.%f%z;
        - %y-%m-%dT%H:%M:%S.%f%z, %y-%m-%d %H:%M:%S.%f%z, %y-%m-%d-%H:%M:%S.%f%z;
        - %m/%d/%yT%H:%M:%S.%f%z, %m/%d/%y %H:%M:%S.%f%z, %m/%d/%y-%H:%M:%S.%f%z;
        - %d/%m/%yT%H:%M:%S.%f%z, %d/%m/%y %H:%M:%S.%f%z, %d/%m/%y-%H:%M:%S.%f%z;
        - %y/%m/%dT%H:%M:%S.%f%z, %y/%m/%d %H:%M:%S.%f%z, %y/%m/%d-%H:%M:%S.%f%z;
        - %m.%d.%yT%H:%M:%S.%f%z, %m.%d.%y %H:%M:%S.%f%z, %m.%d.%y-%H:%M:%S.%f%z;
        - %d.%m.%yT%H:%M:%S.%f%z, %d.%m.%y %H:%M:%S.%f%z, %d.%m.%y-%H:%M:%S.%f%z;
        - %y.%m.%dT%H:%M:%S.%f%z, %y.%m.%d %H:%M:%S.%f%z, %y.%m.%d-%H:%M:%S.%f%z;
        """

        date_text = self.df[column].dropna()

        n_samples = min(100, len(date_text))
        sample = date_text.sample(n_samples).values

        types = [self.__get_date_format(str(i)) for i in sample]

        if not any(types):
            return "%d-%m-%Y"

        most_common = Counter(types).most_common()
        if most_common[0][0] is None:
            chosen_format = most_common[1][0] if len(most_common) > 1 else "%d-%m-%Y"
        else:
            chosen_format = most_common[0][0]

        return chosen_format

    @staticmethod
    def __get_date_format(date_string: str) -> str:
        """
        Get a date format and remove timezone abbreviation, if present
        """
        date_format = guess_datetime_format(date_string)
        if not date_format:
            return None

        match = TIMEZONE_REGEX.search(date_format)
        if match and (abbr := match.group("tz_abbr")):
            date_format = date_format.replace(abbr, "%Z")

        return date_format

    def _set_date_format(self):
        """
        Define the date format for each date column
        """
        self.date_mapping = {
            column: self.__define_date_format(column)
            for column in self.date_columns
        }

    def _csv_data_pipeline(self):
        """
        Divide columns in dataframe into groups -
        binary, categorical, integer, float, string, date
        in case metadata of the table is absent
        """
        for col in self.df.columns:
            col_no_na = self.df[col].dropna()

            if col_no_na.dtype in ["int", "int64"] or col in self.cast_to_integer:
                self.int_columns.add(col)
            elif col_no_na.dtype in ["float", "float64"] or col in self.cast_to_float:
                self.float_columns.add(col)

        float_to_int_cols = set()
        for col in self.float_columns:
            if all(x.is_integer() for x in self.df[col].dropna()):
                float_to_int_cols.add(col)

        self.int_columns = (self.int_columns | float_to_int_cols) - (
            self.categorical_columns | self.binary_columns
        )
        self.float_columns = (
            self.float_columns - self.categorical_columns - self.int_columns - self.binary_columns
        )
        self.str_columns = (
            set(self.df.columns)
            - self.float_columns
            - self.categorical_columns
            - self.int_columns
            - self.binary_columns
            - self.long_text_columns
            - self.email_columns
            - self.uuid_columns
        )
        self.categorical_columns -= self.long_text_columns
        self._set_date_columns()
        self.str_columns -= self.date_columns
        self.uuid_columns = self.uuid_columns - self.categorical_columns - self.binary_columns
        self.uuid_columns_types = {
            k: v for k, v in self.uuid_columns_types.items() if k in self.uuid_columns
        }
        self._set_date_format()

    def _avro_data_pipeline(self):
        """
        Divide columns in dataframe into groups - binary, categorical, integer, float, string, date
        in case metadata of the table in Avro format is present
        """
        self.int_columns = set(
            column for column, data_type in self.fields.items() if data_type == "int"
        )
        self.int_columns = (self.int_columns - self.categorical_columns - self.binary_columns)
        self.float_columns = set(
            column for column, data_type in self.fields.items() if data_type == "float"
        )
        self.float_columns = self.float_columns - self.categorical_columns - self.binary_columns
        self.str_columns = set(
            column for column, data_type in self.fields.items() if data_type == "string"
        )
        self.categorical_columns -= self.long_text_columns
        self.str_columns = (
            self.str_columns
            - self.categorical_columns
            - self.binary_columns
            - self.long_text_columns
            - self.email_columns
            - self.uuid_columns
        )
        self._set_date_columns()
        self.str_columns -= self.date_columns
        self.uuid_columns = self.uuid_columns - self.categorical_columns - self.binary_columns
        self.uuid_columns_types = {
            k: v for k, v in self.uuid_columns_types.items() if k in self.uuid_columns
        }
        self._set_date_format()

    def _detection_pipeline(self):
        if self.schema_format == "CSV":
            self._csv_data_pipeline()
        elif self.schema_format == "Avro":
            self._avro_data_pipeline()

        assert (
            len(self.str_columns) + len(self.float_columns) + len(self.int_columns)
            + len(self.date_columns) + len(self.categorical_columns) + len(self.binary_columns)
            + len(self.long_text_columns) + len(self.uuid_columns) + len(self.email_columns)
        ) == len(self.df.columns), (
            "According to number of columns with defined types, "
            "column types are not identified correctly"
        )

        logger.debug(
            f"Count of string columns: {len(self.str_columns)}; "
            + f"Count of email columns: {len(self.email_columns)}; "
            + f"Count of float columns: {len(self.float_columns)}; "
            + f"Count of int columns: {len(self.int_columns)}; "
            + f"Count of categorical columns: {len(self.categorical_columns)}; "
            + f"Count of date columns: {len(self.date_columns)}; "
            + f"Count of binary columns: {len(self.binary_columns)}; "
            + f"Count of long text columns: {len(self.long_text_columns)}; "
            + f"Count of uuid columns: {len(self.uuid_columns)}"
        )

    def assign_feature(self, feature, columns):
        name = feature.original_name

        if name in self.features:
            raise Exception(f"{name} is already contained in features")

        if not isinstance(columns, (list, tuple)):
            columns = [columns]

        self.features[name] = feature
        self.columns[name] = columns

    def fit(self):
        for name, feature in self.features.items():
            feature.fit(
                self.df[self.columns[name]],
                date_mapping=self.date_mapping,
                to_datetime_conversion=self.to_datetime_conversion
            )

        self.all_columns = [col for col in self.columns]
        self.is_fitted = True

    def transform(self, data, excluded_features=set()):
        transformed_features = list()
        selected_features = {
            name: feature
            for name, feature in self.features.items()
            if name not in excluded_features
        }
        for name, feature in selected_features.items():
            transformed_features.append(feature.transform(data[self.columns[name]]))
        return transformed_features

    def fit_transform(self, data):
        self.fit()
        return self.transform(data)

    def _check_count_features(self, data):
        return (len(data) == len(self.features)) or (
            len(data) + len(self.fk_columns) == len(self.features)
        )

    def inverse_transform(self, data, excluded_features=set()):
        inverse_transformed_data = list()
        column_names = list()
        if not isinstance(data, list):
            data = [data]
        assert self._check_count_features(data)

        for transformed_data, (name, feature) in tqdm.tqdm(
            iterable=zip(data, self.features.items()),
            desc="Generation of the data...",
            total=len(data),
        ):
            if name not in excluded_features and name not in self.fk_columns:
                column_names.extend(self.columns[name])
                logger.trace(f'Column {name} is being inverse transformed.')
                inverse_transformed_data.append(feature.inverse_transform(transformed_data))

        stacked_data = np.column_stack(inverse_transformed_data)
        data = pd.DataFrame(stacked_data, columns=column_names)

        return data

    def _preprocess_str_params(self, feature: str) -> Tuple[int, int]:
        max_len = int(self.df[feature].apply(lambda line: len(line)).max())
        rnn_units = 16
        if 1 <= max_len < 7:
            rnn_units = 32

        if 6 < max_len < 13:
            rnn_units = 128

        if 12 < max_len < 17:
            rnn_units = 256

        if max_len > 16:
            rnn_units = 512
        return max_len, rnn_units

    def _preprocess_nan_cols(
        self, feature: str, fillna_strategy: str = None, zero_cutoff: float = 0.3
    ) -> tuple:
        """Fill NaN values in numeric column with some value according to strategy.
        Fill NaN values in string columns can only work in 'mode' strategy.
        If NaN values exist additional column is created and added to DataFrame.
        This column has value of 1 in case corresponding row contains NaN and 0 otherwise.
        New column name is built like 'column name'+'_null'.

        Args:
            feature (str): Feature name.
            fillna_strategy (str, optional): Can be 'mean', 'mode' or None.
                                             If None NaN values in column are replaced with 0.
                                             Defaults to None.
                                             Note: string columns only work with 'mode'.

        Returns:
            tuple: Tuple that consists of either feature name or both feature name
            and new null feature name.
        """
        isnull_feature = pd.isnull(self.df[feature])
        num_zeros = (self.df[feature] == 0).sum()
        many_zeros_feature = (num_zeros / (len(self.df[feature]))) > zero_cutoff
        if many_zeros_feature:
            feature_zero = feature + "_zero"
            self.df[feature_zero] = self.df[feature].apply(lambda x: 0 if x == 0 else 1)
            if not isnull_feature.any():
                logger.info(
                    f"Column '{feature}' contains {num_zeros} "
                    f"({round(num_zeros * 100 / (len(self.df[feature])))}%) "
                    f"zero values out of {(len(self.df[feature]))}. "
                    f"Ratio of zeros will be preserved in synthetic data."
                )
                return (feature, feature_zero)
        if isnull_feature.any():
            nan_number = isnull_feature.sum()
            logger.info(
                f"Column '{feature}' contains {nan_number} "
                f"({round(nan_number * 100 / len(isnull_feature), 2)}%) "
                f"empty values out of {len(isnull_feature)}. "
                f"Filling them with {fillna_strategy or 'zero'}."
            )
            if fillna_strategy == "mean":
                fillna_value = self.df[feature].mean()
            elif fillna_strategy == "mode":
                fillna_value = self.df[feature].dropna().mode().sample(1).values[0]
            elif (fillna_strategy == "text") or (fillna_strategy == "email"):
                fillna_value = ""
            else:
                fillna_value = 0

            feature_null = feature + "_null"
            self.df[feature_null] = isnull_feature.astype(int)
            self.df[feature] = self.df[feature].fillna(fillna_value)
            if not many_zeros_feature:
                return (feature, feature_null)
            else:
                return (feature, feature_null, feature_zero)
        else:
            return (feature,)

    def _preprocess_categ_params(self, feature: str, strategy: Literal["?", "fill"] = "?"):
        """
        Preprocess categorical columns by filling NaN values with a strategy
        """
        if self.df[feature].isnull().any():
            if strategy == "?":
                self.df[feature] = self.df[feature].fillna("?").astype(str)
            if strategy == "fill":
                self.df[feature] = self.df[feature].fillna(method="bfill").fillna(method="ffill")
                logger.info(
                    f"Filling NaN values in column - '{feature}' with 'bfill' and 'ffill' methods."
                )
        return feature

    @staticmethod
    @slugify_parameters(exclude_params=("fk_kde_path", "fk_column"))
    def _fetch_mapper(fk_kde_path, table_name, pk_table, pk_column, fk_column):
        """
        Fetch the mapper for foreign key in data type - 'string'
        """
        try:
            with open(
                f"{fk_kde_path.replace(table_name, pk_table)}{pk_column}_mapper.pkl",
                "rb",
            ) as file:
                mapper = pickle.load(file)
            return mapper
        except FileNotFoundError:
            logger.warning(
                f"The mapper for the {fk_column} text key is not found. "
                f"Simple sampling will be used."
            )

    @staticmethod
    @slugify_parameters(exclude_params=("kde", "fk_kde_path"))
    def _save_kde_artifacts(kde, fk_kde_path, fk_column):
        """
        Save KDE artifacts
        """
        with open(f"{fk_kde_path}{fk_column}.pkl", "wb") as file:
            dill.dump(kde, file)

        logger.info(f"KDE artifacts saved to {fk_kde_path}{fk_column}.pkl")

    def _preprocess_fk_params(self):
        for fk in self.foreign_keys_list:
            fk_columns = self.foreign_keys_mapping.get(fk).get("columns")
            for fk_column in fk_columns:
                fk_column_values = self.df[fk_column]
                correspondent_pk_table = self.foreign_keys_mapping[fk]["references"]["table"]
                correspondent_pk_col = self.foreign_keys_mapping[fk]["references"]["columns"][0]
                if fk_column_values.dtype in (pd.StringDtype(), "object"):
                    mapper = self._fetch_mapper(
                        fk_kde_path=self.paths["fk_kde_path"],
                        table_name=self.table_name,
                        pk_table=correspondent_pk_table,
                        pk_column=correspondent_pk_col,
                        fk_column=fk_column,
                    )
                    if mapper is None:
                        continue
                    fk_column_values = fk_column_values.map(mapper)
                noise_to_prevent_singularity = np.random.normal(0, 0.0001, len(fk_column_values))
                kde = gaussian_kde(fk_column_values + noise_to_prevent_singularity)
                self._save_kde_artifacts(
                    kde=kde, fk_kde_path=self.paths["fk_kde_path"], fk_column=fk_column
                )

    def _drop_fk_columns(self):
        """
        Drop columns in dataframe which defined as foreign key
        """
        for fk_column in set(self.fk_columns):
            self.df = self.df.drop(fk_column, axis=1)
            logger.debug(
                f"The column - '{fk_column}' dropped from the training process "
                f"as it is defined as FK column and will be sampled from the PK table"
            )

    def _assign_char_feature(self, feature):
        """
        Assign text based feature to text columns
        """
        features = self._preprocess_nan_cols(feature, fillna_strategy="text")
        max_len, rnn_units = self._preprocess_str_params(features[0])
        self.assign_feature(
            CharBasedTextFeature(features[0], text_max_len=max_len, rnn_units=rnn_units),
            features[0],
        )
        logger.info(f"Column '{features[0]}' assigned as text based feature")

        if len(features) > 1:
            for feature in features[1:]:
                if feature.endswith("_null"):
                    self.null_num_column_names.append(feature)
                if feature.endswith("_zero"):
                    self.zero_num_column_names.append(feature)
                self.assign_feature(ContinuousFeature(feature, column_type=float), feature)
                logger.info(f"Column '{feature}' assigned as float based feature")

    def _assign_email_feature(self, feature):
        """
        Assign email feature to text columns
        """
        features = self._preprocess_nan_cols(feature, fillna_strategy="email")
        max_len, rnn_units = 15, 32
        self.assign_feature(
            EmailFeature(features[0], text_max_len=max_len, rnn_units=rnn_units),
            features[0],
        )
        logger.info(f"Column '{features[0]}' assigned as email feature")

        # TODO: encapsulate this logic in a separate function
        if len(features) > 1:
            for feature in features[1:]:
                if feature.endswith("_null"):
                    self.null_num_column_names.append(feature)
                if feature.endswith("_zero"):
                    self.zero_num_column_names.append(feature)
                self.assign_feature(ContinuousFeature(feature, column_type=float), feature)
                logger.info(f"Column '{feature}' assigned as float based feature")

    def _assign_float_feature(self, feature):
        """
        Assign float based feature to float columns
        """
        # num_bins = self.find_clusters(df, float_columns)
        features = self._preprocess_nan_cols(feature, fillna_strategy="mean")
        if len(features) == 2 and features[1].endswith("_null"):
            self.null_num_column_names.append(features[1])
        if len(features) == 2 and features[1].endswith("_zero"):
            self.zero_num_column_names.append(features[1])
        if len(features) == 3:
            self.null_num_column_names.append(features[1])
            self.zero_num_column_names.append(features[2])
        for feature in features:
            self.assign_feature(ContinuousFeature(feature, column_type=float), feature)
            logger.info(f"Column '{feature}' assigned as float based feature")

    def _assign_int_feature(self, feature):
        """
        Assign int based feature to int columns
        """
        features = self._preprocess_nan_cols(feature, fillna_strategy="mean")
        self.assign_feature(ContinuousFeature(features[0], column_type=int), features[0])
        logger.info(f"Column '{features[0]}' assigned as int based feature")
        if len(features) > 1:
            for feature in features[1:]:
                if feature.endswith("_null"):
                    self.null_num_column_names.append(feature)
                if feature.endswith("_zero"):
                    self.zero_num_column_names.append(feature)
                self.assign_feature(ContinuousFeature(feature, column_type=float), feature)
                logger.info(f"Column '{feature}' assigned as float based feature")

    def _assign_categ_feature(self, feature, strategy: Literal["?", "fill"] = "?"):
        """
        Assign categorical based feature to categorical columns
        """
        feature = self._preprocess_categ_params(feature, strategy)
        self.assign_feature(CategoricalFeature(feature), feature)
        logger.info(f"Column '{feature}' assigned as categorical based feature.")

    def _preprocess_dates_with_timezone(self, feature):
        """
        Preprocess date columns with timezone information,
        adding a new column with timezone information if applicable.
        """
        timezone_data = self.df[feature].map(lambda x: fetch_timezone(x))
        if timezone_data.isnull().all():
            return

        self.df[f"{feature}_tz"] = timezone_data
        percent_with_tz = round(timezone_data.notnull().mean() * 100, 2)
        unique_tz = timezone_data.dropna().unique()
        unique_tz = ', '.join(
            unique_tz[:5].tolist() + ["etc"] if len(unique_tz) > 5 else unique_tz
        )

        logger.info(
            f"Column '{feature}' contains {percent_with_tz}% dates with time zone. "
            f"Unique time zones: {unique_tz}."
        )
        self._assign_categ_feature(f"{feature}_tz", strategy="fill")

    def _assign_date_feature(self, feature):
        """
        Assign date feature to date columns
        """
        date_format = self.date_mapping.get(feature, str())
        if "%z" in date_format.lower():
            self._preprocess_dates_with_timezone(feature)
        features = self._preprocess_nan_cols(feature, fillna_strategy="mode")
        self.assign_feature(DateFeature(features[0]), features[0])
        logger.info(f"Column '{features[0]}' assigned as date feature.")
        if len(features) == 2:
            self.null_num_column_names.append(features[1])
            self.assign_feature(ContinuousFeature(features[1]), features[1])
            logger.info(f"Column '{features[1]}' assigned as float feature.")

    def _assign_binary_feature(self, feature):
        """
        Assign binary feature to binary columns
        """
        feature = self._preprocess_categ_params(feature)
        self.assign_feature(BinaryFeature(feature), feature)
        logger.info(f"Column '{feature}' assigned as binary feature")

    def _assign_fk_feature(self):
        """
        Assign corresponding to FK null column and preprocess if required.
        """
        for fk_column in self.fk_columns:
            features = self._preprocess_nan_cols(fk_column, fillna_strategy="mode")
            if len(features) > 1:
                self.assign_feature(
                    ContinuousFeature(features[1], column_type=int), features[1]
                )

    def _assign_uuid_null_feature(self, feature):
        """
        Assign corresponding to uuid column null column and preprocess if required.
        """
        logger.info(f"Column '{feature}' defined as UUID column")
        features = self._preprocess_nan_cols(feature, fillna_strategy="text")
        if len(features) == 2:
            self.null_num_column_names.append(features[1])
            self.assign_feature(ContinuousFeature(features[1]), features[1])
            logger.info(f"Column '{features[1]}' assigned as float feature")

    def __prepare_primary_key_mapping(self):
        self.primary_keys_mapping.update(self.unique_keys_mapping)
        pk_uq_keys_mapping = self.primary_keys_mapping
        if pk_uq_keys_mapping:
            self.__set_types(pk_uq_keys_mapping)
            self.__map_text_pk()

    def _assign_feature(self, column: str):
        """
        Assign feature to the column based on its type
        """
        for column_type, handler in self.dispatcher.items():
            if column in getattr(self, column_type, []):
                handler(column)
                return

    @property
    def dispatcher(self):
        """
        Dispatcher for assigning features to columns
        """
        return {
            "str_columns": self._assign_char_feature,
            "email_columns": self._assign_email_feature,
            "float_columns": self._assign_float_feature,
            "int_columns": self._assign_int_feature,
            "categorical_columns": self._assign_categ_feature,
            "date_columns": self._assign_date_feature,
            "binary_columns": self._assign_binary_feature,
            "uuid_columns": self._assign_uuid_null_feature,
        }

    def __assign_features(self):
        """
        Assign features to the columns based on their types
        """
        for column in [col for col in self.df.columns]:
            self._assign_feature(column)

    def _ensure_technical_column_if_no_features(self, additional_message: str = ""):
        """
        Workaround for the case when all columns are dropped.
        Add a technical column to proceed with the training process.
        """
        if not self.features:
            tech_column = "syngen_tech_column"
            logger.info(
                f"Since all columns in the table '{self.table_name}' "
                f"are uuid/key/long text{additional_message} columns, "
                "there are no suitable columns to train on. "
                f"A technical column '{tech_column}' will be added "
                "to proceed with the training process "
                "and will be removed afterwards."
            )
            self.df[tech_column] = 1
            self._assign_float_feature(tech_column)
            self.tech_columns.add(tech_column)

    def pipeline(self) -> pd.DataFrame:
        if self.foreign_keys_list:
            self._assign_fk_feature()
            self._preprocess_fk_params()
            self._drop_fk_columns()

        self.__prepare_primary_key_mapping()

        self.__assign_features()
        self._ensure_technical_column_if_no_features()

        self.fit()

        # The end of the run related to the preprocessing stage
        MlflowTracker().end_run()

        return self.df



from itertools import chain
from typing import Union, List
from lazy import lazy
from loguru import logger

import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow.keras.backend as K
from scipy.stats import shapiro, kurtosis
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
    QuantileTransformer,
    OneHotEncoder
)
from tensorflow.keras import losses
from tensorflow.keras.layers import (
    Bidirectional,
    Dense,
    Input,
    LSTM,
    Layer,
    RepeatVector,
    TimeDistributed,
)

from syngen.ml.utils import (
    slugify_parameters,
    inverse_dict,
    datetime_to_timestamp,
    convert_to_date
)

KURTOSIS_THRESHOLD = 50  # threshold for kurtosis to consider extreme outliers


class BaseFeature:
    """
    Base class for feature classes.
    Each feature class implements feature preprocessing, transformation and inverse transformation.
    What is more, each feature class contains modules for the neural network (NN), including
    corresponding input, encoder, decoder, and loss
    """

    def __init__(self, name):
        self.name: str = self._reset_name(name=name)
        self.original_name: str = name
        self.weight: float = 1.0

    @staticmethod
    @slugify_parameters(turn_on=False)
    def _reset_name(name):
        """
        Slugify the attribute 'name' of the instance
        """
        return name

    def fit(self, data: pd.DataFrame, **kwargs):
        """
        Fit scalars, one-hot encoders, and mappers in the data
        """
        pass

    def transform(self, data: pd.DataFrame) -> List:
        """
        Transform feature to numeric format according to the fitted preprocessors
        """
        pass

    def inverse_transform(self, data: List) -> np.ndarray:
        """
        Inverse transform feature from numeric to original format to obtain an original-like table
        """
        pass

    def input(self) -> tf.Tensor:
        """
        Define a feature-specific input for the NN
        """
        pass

    def encoder(self) -> tf.Tensor:
        """
        Define a feature-specific encoder for the NN
        """
        pass

    def __decoder_layer(self) -> tf.Tensor:
        """
        Define an elementary layer for decoder to use in create_decoder() method
        """
        pass

    def create_decoder(self, encoder_output: tf.Tensor):
        """
        Create a feature-specific decoder combining given decoder layers and encoder outputs
        """
        pass

    def loss(self) -> tf.Tensor:
        """
        Define a feature-specific loss taking into account the data types
        """
        pass


class BinaryFeature(BaseFeature):
    """
    A class to process binary features, i.e. features containing only two unique values
    """

    def __init__(self, name: str):
        super().__init__(name=name)
        self.feature_type = "categorical"

    def fit(self, data: pd.DataFrame, **kwargs):
        self.mapping = {k: n for n, k in enumerate(np.unique(data))}
        self.inverse_mapping = inverse_dict(self.mapping)
        self.inverse_vectorizer = np.vectorize(self.inverse_mapping.get)
        self.input_dimension = data.shape[1]

    def transform(self, data: pd.DataFrame):
        data = data.astype("object").replace(self.mapping)
        data = data.replace(self.mapping)
        return data.astype("float32")

    def inverse_transform(self, data: List) -> np.ndarray:
        data = np.round(data)
        inversed = self.inverse_vectorizer(data)
        return np.where(inversed == "?", None, inversed)

    @lazy
    def input(self) -> tf.Tensor:
        return Input(shape=(self.input_dimension,), name="input_%s" % self.name)

    @lazy
    def encoder(self) -> tf.Tensor:
        return self.input

    @lazy
    def __decoder_layer(self) -> tf.Tensor:
        return Dense(self.input_dimension, activation="sigmoid")

    def create_decoder(self, encoder_output: tf.Tensor) -> tf.Tensor:
        self.decoder = self.__decoder_layer(encoder_output)
        return self.decoder

    @lazy
    def loss(self) -> tf.Tensor:
        if not hasattr(self, "decoder"):
            Exception("Decoder isn't created")

        return self.weight * losses.binary_crossentropy(self.input, self.decoder)


class ContinuousFeature(BaseFeature):
    """
    A class to process the continuous numeric features, including floats and integers
    """

    def __init__(
        self,
        name: str,
        decoder_layers: Union[None, tuple, int] = (60,),
        weight_randomizer: Union[None, bool, tuple] = None,
        column_type=float,
    ):
        super().__init__(name=name)
        if decoder_layers is None:
            decoder_layers = ()
        elif isinstance(decoder_layers, int):
            decoder_layers = (decoder_layers,)

        # TODO: teset features
        if weight_randomizer is None or (
            isinstance(weight_randomizer, bool) and not weight_randomizer
        ):
            weight_randomizer = (1, 1)
        elif isinstance(weight_randomizer, bool) and weight_randomizer:
            weight_randomizer = (0, 1)
        elif isinstance(weight_randomizer, (float, int)):
            weight_randomizer = (weight_randomizer, weight_randomizer)

        self.decoder_layers = decoder_layers
        self.weight_randomizer = weight_randomizer
        self.column_type = column_type
        self.feature_type = "numeric"

    def fit(self, data: pd.DataFrame, **kwargs):
        self.is_positive = (data >= 0).sum().item() >= len(data) * 0.99
        self.scaler = self._select_scaler(data)
        self.scaler.fit(data)
        self.input_dimension = data.shape[1]

    def transform(self, data: pd.DataFrame) -> np.ndarray:
        return self.scaler.transform(data).astype("float32")

    def inverse_transform(self, data: np.ndarray) -> np.ndarray:
        reverse_transformed = self.scaler.inverse_transform(data)
        reverse_transformed = (
            np.abs(reverse_transformed) if self.is_positive else reverse_transformed
        )
        return (
            reverse_transformed
            if self.column_type is float
            else np.around(reverse_transformed).astype("int64")
        )

    def _select_scaler(
            self, data: pd.DataFrame,
            kurtosis_threshold=KURTOSIS_THRESHOLD
    ) -> object:
        """
        Select appropriate scaler based on data characteristics.

        Strategy:
        - Extreme outliers -> QuantileTransformer
        - Normal distribution -> StandardScaler
        - "_null" features and other cases -> MinMaxScaler
        """
        data = data.iloc[:, 0]

        kurt = kurtosis(data)
        normality = shapiro(data.sample(n=min(len(data), 500))).pvalue

        if normality >= 0.05:
            return StandardScaler()

        # Column that ends with "_null"
        # (indicators of nan values of corresponding column)
        # is binary but stored as numeric.
        # It must not be transformed with QuantileTransformer
        if data.nunique() == 2:
            return MinMaxScaler()

        # QuantileTransformer for extreme outliers
        if kurt > kurtosis_threshold:
            logger.debug(
                f"Column '{self.name}' has extreme outliers: "
                f"kurtosis={kurt:.2f} > kurtosis_threshold={kurtosis_threshold}. "
                f"Using QuantileTransformer."
            )
            quantile_params = self._get_quantile_transformer_params(
                n_samples=len(data),
                kurt=kurt,
                kurtosis_threshold=kurtosis_threshold
            )
            return QuantileTransformer(**quantile_params)
        else:
            return MinMaxScaler()

    def _get_quantile_transformer_params(
            self, n_samples, kurt, kurtosis_threshold
            ) -> dict:
        """
        Get optimal parameters for QuantileTransformer
        based on data characteristics
        """
        base_quantiles = min(n_samples, 100_000)

        # for distributions with very extreme outliers use more quantiles
        if kurt > 4 * kurtosis_threshold:
            quantile_factor = 1.5
        elif kurt > 2 * kurtosis_threshold:
            quantile_factor = 1.2
        else:
            quantile_factor = 1.0

        n_quantiles = min(int(base_quantiles * quantile_factor), n_samples)

        # supsample did not affect time for fitting and transforming
        subsample = None

        return {
            'n_quantiles': n_quantiles,
            'subsample': subsample,
            'output_distribution': 'normal'
        }

    @lazy
    def input(self) -> tf.Tensor:
        return Input(shape=(self.input_dimension,), name="input_%s" % self.name)

    @lazy
    def encoder(self) -> tf.Tensor:
        return self.input

    @lazy
    def __decoder_layer(self) -> List[tf.Tensor]:
        decoder_layers = list()
        for idx, item in enumerate(self.decoder_layers):
            name = "%s_decoder_%d" % (self.name, idx)
            if isinstance(item, int):
                decoder_layers.append(Dense(item, activation="relu", name=name))

            if isinstance(item, Layer):
                item.name = name
                decoder_layers.append(Layer)

        decoder_layers.append(
            Dense(self.input_dimension, activation="linear", name="%s_linear" % self.name)
        )

        return decoder_layers

    def create_decoder(self, encoder_output: tf.Tensor) -> tf.Tensor:
        if not isinstance(self.__decoder_layer, list):
            decoder_layers = [self.__decoder_layer]
        else:
            decoder_layers = self.__decoder_layer

        x = encoder_output
        for layer in decoder_layers:
            x = layer(x)

        self.decoder = x
        return self.decoder

    @lazy
    def loss(self) -> tf.Tensor:
        if not hasattr(self, "decoder"):
            Exception("Decoder isn't created")

        low = self.weight_randomizer[0]
        high = self.weight_randomizer[1]
        random_weight = K.random_uniform_variable(shape=(1,), low=low, high=high)

        return random_weight * tf.keras.losses.MSE(self.input, self.decoder)


class CategoricalFeature(BaseFeature):
    """
    A class to process categorical values, i.e. values with 2 < unique_values < 50
    """

    def __init__(
        self,
        name: str,
        decoder_layers: Union[None, tuple, int] = (60,),
        weight_randomizer: Union[None, bool, tuple] = None,
    ):
        if decoder_layers is None:
            decoder_layers = ()
        elif isinstance(decoder_layers, int):
            decoder_layers = (decoder_layers,)

        if weight_randomizer is None or (
            isinstance(weight_randomizer, bool) and not weight_randomizer
        ):
            weight_randomizer = (1, 1)
        elif isinstance(weight_randomizer, (float, int)):
            weight_randomizer = (weight_randomizer, weight_randomizer)
        elif isinstance(weight_randomizer, bool) and weight_randomizer:
            weight_randomizer = (0, 1)

        super().__init__(name="_".join(name.split()))

        self.one_hot_encoder = OneHotEncoder(
            sparse_output=False,
            handle_unknown='ignore',
            dtype=np.float32)
        self.decoder = None
        self.decoder_layers = decoder_layers
        self.weight_randomizer = weight_randomizer
        self.feature_type = "categorical"

    def fit(self, data: pd.DataFrame, **kwargs):
        """
        Fit the encoder and create mappings.
        """
        data = data.iloc[:, 0].astype(str).to_numpy().reshape(-1, 1)

        self.one_hot_encoder.fit(data)

        categories = self.one_hot_encoder.categories_[0]
        self.mapping = {cat: idx for idx, cat in enumerate(categories)}
        self.inverse_mapping = inverse_dict(self.mapping)
        self.inverse_vectorizer = np.vectorize(self.inverse_mapping.get)

        self.input_dimension = len(self.mapping)

    def transform(self, data: pd.DataFrame) -> np.ndarray:
        """
        Transform data to one-hot encoding.
        """
        data = data.iloc[:, 0].astype(str).to_numpy().reshape(-1, 1)

        return self.one_hot_encoder.transform(data).astype("float32")

    def inverse_transform(self, data: np.ndarray) -> np.ndarray:
        """
        Convert one-hot encoded data back to original categories.
        """
        data = data.argmax(axis=1)
        inversed = self.inverse_vectorizer(data)

        return np.where(inversed == "?", None, inversed)

    @lazy
    def input(self) -> tf.Tensor:
        self.idx_input = Input(shape=(self.input_dimension,), name="input_%s" % self.name)

        return self.idx_input

    @lazy
    def encoder(self) -> tf.Tensor:
        return self.idx_input

    @lazy
    def __decoder_layer(self) -> List[tf.Tensor]:
        decoder_layers = list()
        for idx, item in enumerate(self.decoder_layers):
            name = "%s_decoder_%d" % (self.name, idx)
            if isinstance(item, int):
                decoder_layers.append(Dense(item, activation="relu", name=name))

            if isinstance(item, Layer):
                item.name = name
                decoder_layers.append(Layer)

        decoder_layers.append(
            Dense(
                self.input_dimension,
                activation="softmax",
                name="%s_softmax" % self.name,
            )
        )

        return decoder_layers

    def create_decoder(self, encoder_output: tf.Tensor) -> tf.Tensor:
        if not isinstance(self.__decoder_layer, list):
            decoder_layers = [self.__decoder_layer]
        else:
            decoder_layers = self.__decoder_layer

        x = encoder_output
        for layer in decoder_layers:
            x = layer(x)

        self.decoder = x
        return self.decoder

    @lazy
    def loss(self) -> tf.Tensor:
        if not hasattr(self, "decoder"):
            Exception("Decoder isn't created")

        low = self.weight_randomizer[0]
        high = self.weight_randomizer[1]
        random_weight = K.random_uniform_variable(shape=(1,), low=low, high=high)

        return random_weight * tf.keras.losses.categorical_crossentropy(self.input, self.decoder)


class CharBasedTextFeature(BaseFeature):
    """
    A class to process the text features
    """

    def __init__(
        self,
        name: str,
        text_max_len: int,
        rnn_units: int = 128,
        dropout: int = 0,
    ):
        super().__init__(name=name)
        self.decoder = None
        self.text_max_len = text_max_len
        self.rnn_units = rnn_units
        self.rnn_unit = LSTM
        self.dropout = dropout
        self.feature_type = "text"

    def fit(self, data: pd.DataFrame, **kwargs):
        from tensorflow.keras.preprocessing.text import Tokenizer

        if len(data.columns) > 1:
            raise Exception("CharBasedTextFeature can work only with one text column")

        data = data[data.columns[0]]

        tokenizer = Tokenizer(lower=False, char_level=True)
        tokenizer.fit_on_texts(data)
        tokenizer.inverse_dict = inverse_dict(tokenizer.word_index)

        self.vocab_size = len(tokenizer.word_index)
        self.tokenizer = tokenizer

    def transform(self, data: pd.DataFrame) -> np.ndarray:
        if len(data.columns) > 1:
            raise Exception("CharBasedTextFeature can work only with one text column")

        data = data[data.columns[0]]

        from tensorflow.keras.preprocessing.sequence import pad_sequences

        data_gen = self.tokenizer.texts_to_sequences(data)
        data_gen = pad_sequences(
            data_gen,
            maxlen=self.text_max_len,
            padding="post",
            truncating="post",
            value=0.0,
        )
        # return data_gen
        return K.one_hot(K.cast(data_gen, "int32"), self.vocab_size)

    @staticmethod
    def _top_p_filtering(
            logits: np.ndarray,
            top_p: float = 0.9
    ):
        # Convert logits to TensorFlow tensor
        logits = tf.convert_to_tensor(logits, dtype=tf.float32)

        # Sort logits and get sorted indices
        sorted_logits = tf.sort(logits, direction="DESCENDING", axis=-1)
        sorted_indices = tf.argsort(logits, direction="DESCENDING", axis=-1)

        # Calculate cumulative probabilities
        cumulative_probs = tf.cumsum(sorted_logits, axis=-1)

        # Remove tokens with cumulative probability above the threshold
        sorted_indices_to_remove = cumulative_probs >= top_p

        # Shift the indices to the right to keep also the first token above the threshold
        zeros_for_shift = tf.zeros_like(sorted_indices_to_remove[:, :, :1], dtype=tf.bool)
        sorted_indices_to_remove = tf.concat(
            [zeros_for_shift, sorted_indices_to_remove[:, :, :-1]],
            axis=-1
        )

        # Create a mask for indices to remove
        batch_size, seq_length, vocab_size = logits.shape

        batch_indices = tf.repeat(tf.range(batch_size), seq_length * vocab_size)
        feature_length_indices = tf.tile(
            tf.repeat(tf.range(seq_length), vocab_size),
            [batch_size]
        )
        vocab_selection_indices = tf.reshape(sorted_indices, [-1])
        update_indices = tf.stack(
            [batch_indices, feature_length_indices, vocab_selection_indices],
            axis=1
        )
        flattened_update_values = tf.reshape(sorted_indices_to_remove, [-1])
        indices_to_remove = tf.tensor_scatter_nd_update(
            tf.zeros_like(logits, dtype=sorted_indices_to_remove.dtype),
            update_indices,
            flattened_update_values,
        )

        # Apply the filter value to the logits
        logits_removed = tf.where(indices_to_remove, tf.fill(indices_to_remove.shape, 0.0), logits)

        return logits_removed.numpy().astype(np.float64)

    @staticmethod
    def _top_k_filtering(
            logits: np.ndarray,
            top_k: int = 0
    ):
        indices_to_remove = logits < tf.math.top_k(logits, top_k)[0][..., -1, None]
        logits[indices_to_remove] = 0.0
        return logits

    def _process_batch(self, batch: np.ndarray) -> List[str]:
        probs = tf.nn.softmax(batch, axis=-1).numpy().astype(float)
        probs = self._top_p_filtering(probs, top_p=0.9)
        # probs = self._top_k_filtering(probs, top_k=6)
        # TODO: select top_k based on inverse_dict length

        probs /= probs.sum(axis=2, keepdims=True)

        multinomial_samples = np.apply_along_axis(
            lambda x: np.argmax(np.random.multinomial(1, x)), -1, probs
        )

        chars_array = np.vectorize(lambda x: self.tokenizer.inverse_dict.get(x, ''))(
            multinomial_samples)

        # Convert tokens to words
        words = ["".join(sample) for sample in chars_array]

        return words

    def inverse_transform(self, data: np.ndarray, **kwargs) -> List[str]:

        batch_size = 10000
        num_batches = (len(data) + batch_size - 1) // batch_size
        feature_values = []

        for i in range(num_batches):
            batch = data[i * batch_size: (i + 1) * batch_size]
            feature_values.extend(self._process_batch(batch))

        return feature_values

    @lazy
    def input(self) -> tf.Tensor:
        self.index_input = Input(
            shape=(self.text_max_len, self.vocab_size), name="input_%s" % self.name
        )

        return self.index_input

    @lazy
    def encoder(self) -> tf.Tensor:
        rnn_encoder_layer = Bidirectional(self.rnn_unit(self.rnn_units, return_sequences=False))

        rnn_econder = rnn_encoder_layer(self.input)
        return rnn_econder

    @lazy
    def __decoder_layer(self) -> List[tf.Tensor]:
        decoder_layers = list()

        decoder_layers.append(RepeatVector(self.text_max_len))
        decoder_layers.append(self.rnn_unit(self.rnn_units, return_sequences=True))
        decoder_layers.append(TimeDistributed(Dense(self.vocab_size, activation="linear")))

        return decoder_layers

    def create_decoder(self, encoder_output: tf.Tensor) -> tf.Tensor:
        if not isinstance(self.__decoder_layer, list):
            decoder_layers = [self.__decoder_layer]
        else:
            decoder_layers = self.__decoder_layer

        x = encoder_output
        for layer in decoder_layers:
            x = layer(x)

        self.decoder = x
        return self.decoder

    @lazy
    def loss(self) -> tf.Tensor:
        if not hasattr(self, "decoder"):
            Exception("Decoder isn't created")

        return self.weight * K.mean(
            tf.compat.v1.nn.softmax_cross_entropy_with_logits_v2(
                labels=self.input, logits=self.decoder
            )
        )


class EmailFeature(CharBasedTextFeature):
    def __init__(
        self,
        name: str,
        text_max_len: int,
        rnn_units: int = 128,
        dropout: int = 0,
        domain: str = 'tdspora.ai'
    ):
        super().__init__(name=name,
                         text_max_len=text_max_len,
                         rnn_units=rnn_units,
                         dropout=dropout)
        self.domain = domain

    def fit(self, data: pd.DataFrame, **kwargs):
        super().fit(self.extract_email_name(data))

    @staticmethod
    def extract_email_name(data: pd.DataFrame) -> pd.DataFrame:
        pattern = r'^([^@]+).*$'  # returns all the string if there's no "@" in it
        return data.iloc[:, 0].str.extract(pattern).rename({0: data.columns[0]}, axis=1).fillna('')

    def transform(self, data: pd.DataFrame) -> np.ndarray:
        if len(data.columns) > 1:
            raise Exception("CharBasedTextFeature can work only with one text column")

        # TODO: add logic for emails to Dataset._preprocess_nan_cols
        return super().transform(self.extract_email_name(data))

    def inverse_transform(self, data: np.ndarray, **kwargs) -> List[str]:
        return [
            s + "@" +
            self.domain for s in super().inverse_transform(data, **kwargs)
        ]


class DateFeature(BaseFeature):
    """
    A class to process datetime features
    """

    def __init__(self, name, decoder_layers=(60,), weight_randomizer=None):
        if decoder_layers is None:
            decoder_layers = ()
        elif isinstance(decoder_layers, int):
            decoder_layers = (decoder_layers,)

        if weight_randomizer is None or (
            isinstance(weight_randomizer, bool) and not weight_randomizer
        ):
            weight_randomizer = (1, 1)
        elif isinstance(weight_randomizer, bool) and weight_randomizer:
            weight_randomizer = (0, 1)
        elif isinstance(weight_randomizer, (float, int)):
            weight_randomizer = (weight_randomizer, weight_randomizer)

        super().__init__(name=name)
        self.decoder_layers = decoder_layers
        self.weight_randomizer = weight_randomizer
        self.feature_type = "numeric"

    def fit(self, data, **kwargs):
        self.date_format = kwargs["date_mapping"].get(self.original_name)
        self.to_datetime_conversion = kwargs["to_datetime_conversion"].get(self.original_name)
        self.data = chain.from_iterable(data.values)
        self.data = pd.DataFrame(
            list(
                map(
                    lambda d: datetime_to_timestamp(d, self.date_format),
                    self.data
                )
            )
        )
        self.is_positive = (self.data >= 0).sum().item() >= len(self.data) * 0.99
        normality = shapiro(self.data.sample(n=min(len(self.data), 500))).pvalue
        self.data = np.array(self.data).reshape(-1, 1)

        self.scaler = StandardScaler() if normality >= 0.05 else MinMaxScaler()
        self.scaler.fit(self.data)
        self.input_dimension = self.data.shape[1]

    def transform(self, data):
        return self.scaler.transform(self.data)

    def inverse_transform(self, data):
        unscaled = self.scaler.inverse_transform(data)
        unscaled = chain.from_iterable(unscaled)
        return list(
            map(
                lambda t: convert_to_date(
                    value=t,
                    date_format=self.date_format,
                    to_datetime_conversion=self.to_datetime_conversion
                ),
                unscaled,
            )
        )

    @lazy
    def input(self):
        return Input(shape=(self.input_dimension,), name="input_%s" % self.name, dtype="float64")

    @lazy
    def encoder(self):
        return self.input

    @lazy
    def __decoder_layer(self):
        decoder_layers = list()
        for idx, item in enumerate(self.decoder_layers):
            name = "%s_decoder_%d" % (self.name, idx)
            if isinstance(item, int):
                decoder_layers.append(Dense(item, activation="relu", name=name))

            if isinstance(item, Layer):
                item.name = name
                decoder_layers.append(Layer)

        decoder_layers.append(
            Dense(
                self.input_dimension,
                dtype="float32",
                activation="linear",
                name="%s_linear" % self.name,
            )
        )

        return decoder_layers

    def create_decoder(self, encoder_output):
        if not isinstance(self.__decoder_layer, list):
            decoder_layers = [self.__decoder_layer]
        else:
            decoder_layers = self.__decoder_layer

        x = encoder_output
        for layer in decoder_layers:
            x = layer(x)

        self.decoder = x
        return self.decoder

    @lazy
    def loss(self):
        if not hasattr(self, "decoder"):
            Exception("Decoder isn't created")

        low = self.weight_randomizer[0]
        high = self.weight_randomizer[1]
        random_weight = K.random_uniform_variable(shape=(1,), low=low, high=high)

        return random_weight * tf.keras.losses.MSE(self.input, self.decoder)



from typing import Dict, List, Literal, Callable, Optional
import os
from dataclasses import dataclass, field
import json
from collections import defaultdict
import inspect

import pandas as pd
from cryptography.fernet import InvalidToken

from slugify import slugify
from loguru import logger
from syngen.ml.data_loaders import MetadataLoader, DataLoader, DataEncryptor, DataFrameFetcher
from syngen.ml.validation_schema import ValidationSchema, ReportTypes
from syngen.ml.utils import ValidationError, fetch_config, get_source_path_extension


@dataclass
class Validator:
    """
    Class for validating the metadata
    """

    metadata: Dict
    metadata_path: str
    type_of_process: Literal["train", "infer"]
    loader: Optional[Callable[[str], pd.DataFrame]]
    type_of_fk_keys = ["FK"]
    infer_report_types: List[str] = field(
        default_factory=lambda: ReportTypes().infer_report_types
    )
    merged_metadata: Dict = field(default_factory=dict)
    mapping: Dict = field(default_factory=dict)
    existed_columns_mapping: Dict = field(default_factory=dict)
    errors = defaultdict(defaultdict)

    def _define_mapping(self):
        """
        Define the mapping contained the information of the foreign keys
        defined in the metadata
        """
        for table_name, table_metadata in self.metadata.items():
            if table_name == "global":
                continue
            metadata_keys = table_metadata.get("keys", {})
            for key_name, key_data in metadata_keys.items():
                if key_data["type"] not in self.type_of_fk_keys:
                    continue

                self.mapping[key_name] = {
                    "parent_table": key_data["references"]["table"],
                    "parent_columns": key_data["references"]["columns"],
                }

    def _check_conditions(self, metadata: Dict) -> bool:
        """
        Check conditions whether to launch the validation or not
        """
        reports = metadata.get("train_settings", {}).get("reports", [])
        return (
            self.type_of_process == "infer"
            or (
                self.type_of_process == "train" and
                any([item in self.infer_report_types for item in reports])
            )
        )

    def _validate_metadata(self, table_name: str):
        """
        Validate the metadata
        """
        metadata_of_the_table = self.metadata[table_name]
        table_keys = metadata_of_the_table.get("keys", {})
        for key, config in table_keys.items():
            if config["type"] not in self.type_of_fk_keys:
                continue
            self._validate_referential_integrity(
                fk_name=key,
                fk_config=config,
                parent_config=self.merged_metadata[self.mapping[key]["parent_table"]],
            )
            parent_table = self.mapping[key]["parent_table"]
            if parent_table not in self.metadata:
                if self._check_conditions(metadata_of_the_table):
                    self._check_completion_of_training(parent_table)
                    self._check_existence_of_generated_data(parent_table)
                elif self.type_of_process == "train":
                    self._check_completion_of_training(parent_table)
            else:
                continue

    def _validate_schema(self):
        """
        Validate the schema of the metadata
        """
        ValidationSchema(
            metadata=self.metadata,
            validation_of_source=False if self.loader is not None else True,
            process=self.type_of_process
        ).validate_schema()

    def _validate_referential_integrity(self, fk_name: str, fk_config: Dict, parent_config: Dict):
        """
        Validate whether the columns related to the primary key are the same as
        the referenced columns of the foreign key
        """
        keys = parent_config.get("keys", {})
        if not keys or all(config["type"] not in ["PK", "UQ"] for config in keys.values()):
            message = (
                "The information about columns of the primary or unique key "
                f"associated with the columns of the '{fk_config['type']}' key - '{fk_name}' "
                "wasn't found in the metadata of the parent table"
            )
            self.errors["validate referential integrity"][fk_name] = message
            return

        result = any(
            [
                config["columns"] == fk_config["references"]["columns"]
                for config in parent_config.get("keys", {}).values()
                if config["type"] in ["PK", "UQ"]
            ]
        )
        if result is False:
            message = (
                f"The columns of primary or unique key associated with the columns of "
                f"the '{fk_config['type']}' key - '{fk_name}' aren't the same"
            )
            self.errors["validate referential integrity"][fk_name] = message

    def _check_existence_of_generated_data(self, parent_table: str):
        """
        Check if the generated data of the certain parent table exists.
        The generated data is created after the successful execution of the inference process
        of the certain table.
        """
        destination = (
            self.merged_metadata[parent_table].get("infer_settings", {}).get("destination")
        )
        source_extension = get_source_path_extension(
            table_name=parent_table,
            metadata=self.merged_metadata
        )
        if destination is None:
            destination = (
                f"model_artifacts/tmp_store/{slugify(parent_table)}/"
                f"merged_infer_{slugify(parent_table)}{source_extension}"
            )
        if not DataLoader(path=destination).has_existed_path:
            message = (
                f"The generated data of the table - '{parent_table}' hasn't been generated. "
                f"Please, generate the data related to the table - '{parent_table}' first"
            )
            self.errors["check existence of the generated data"][parent_table] = message

    def _get_columns_of_source(
        self,
        path_to_source: str,
        table_name: str,
        **format_settings
    ) -> List[str]:
        """
        Check if the source of the certain table exists, and fetch the list of its columns
        """
        if not DataLoader(path=path_to_source).has_existed_path:
            message = (
                f"It seems that the path to the source of the table - '{table_name}' "
                "isn't correct. Please, check the path to the source of the table - "
                f"'{table_name}'"
            )
            self.errors["check existence of the source"][table_name] = message
        else:
            return DataLoader(path=path_to_source).get_columns(**format_settings)

    def _get_columns_by_loader(self, table_name: str):
        """
        Check whether the callback function provided for loading the data is callable,
        has the correct signature, and fetch the list of columns of the source table
        by calling the `loader`
        """
        # Check if the loader is callable or not
        if not callable(self.loader):
            message = (
                f"The provided loader for the table - '{table_name}' isn't callable. "
                "Please, provide a valid callback function."
            )
            self.errors["check of the loader"][table_name] = message
            return

        # Inspect the signature to ensure it accepts only the parameter `table_name`
        try:
            sig = inspect.signature(self.loader)
            params = sig.parameters
            if "table_name" not in params:
                self.errors["check of the loader"][table_name] = (
                    "The provided loader should accept `table_name` as an argument. "
                    "Please, adjust the signature of the loader."
                )
                return
            if "table_name" in params and len(params) > 1:
                self.errors["check of the loader"][table_name] = (
                    "The provided loader should accept only the argument - `table_name`. "
                    "Please, adjust the signature of the loader."
                )
                return
        except (ValueError, TypeError, AttributeError) as e:
            self.errors["check of the loader"][table_name] = (
                f"Couldn't inspect loader signature: {e}"
            )
            return

        # Fetch the columns by calling the `loader`
        try:
            data_loader = DataFrameFetcher(
                loader=self.loader,
                table_name=table_name
            )
            return data_loader.get_columns()
        except Exception as e:
            self.errors["check of the loader"][table_name] = (
                f"The provided loader raised an exception when called: {e}"
            )
            return

    def _check_existence_of_destination(self, table_name: str):
        """
        Check if the destination of the certain table exists
        """
        destination = self.merged_metadata[table_name].get("infer_settings", {}).get("destination")
        source_extension = get_source_path_extension(
            table_name=table_name,
            metadata=self.merged_metadata
        )
        if destination is None:
            logger.warning(
                f"As the destination path wasn't specified for the table - "
                f"'{table_name}', the synthetic data will be stored "
                f"at the default path - './model_artifacts/tmp_store/{slugify(table_name)}/"
                f"merged_infer_{slugify(table_name)}{source_extension}'"
            )
        if destination is not None and not DataLoader(path=destination).has_existed_destination:
            message = (
                f"It seems that the directory path for storing the generated data of table "
                f"'{table_name}' isn't correct. Please, verify the destination path"
            )
            self.errors["check existence of the destination"][table_name] = message

    def _check_completion_of_training(self, table_name: str):
        """
        Check if the training process of a specific table has been completed.

        Args:
        table_name (str): The name of the table to check.

        Raises:
        FileNotFoundError: If the success file does not exist.
        ValueError: If the content of the success file does not indicate success.
        """
        path_to_success_file = (
            f"model_artifacts/resources/{slugify(table_name)}/train_message.success"
        )
        error_message = (
            f"The training of the table - '{table_name}' hasn't been completed. "
            "Please, retrain the table."
        )

        content = str()
        if os.path.exists(path_to_success_file):
            with open(path_to_success_file, "r") as file:
                content = file.read().strip()

        if content != "SUCCESS":
            self.errors["check completion of the training process"][table_name] = error_message

    def _check_merged_metadata(self, parent_table: str):
        if parent_table not in self.merged_metadata:
            message = (
                f"The metadata of the parent table - '{parent_table}' hasn't been found. "
                f"Please, check whether the information of the parent table - '{parent_table}' "
                f"exists in the current metadata file or in the metadata files stored in "
                f"'model_artifacts/metadata' directory"
            )
            logger.error(message)
            raise ValidationError(message)

    def _merge_metadata(self):
        """
        Find the parent metadata contained the parent table
        in the metadata files stored in 'model_artifacts/metadata' directory,
        and merge it with the metadata of the child table
        """
        self.merged_metadata = self.metadata.copy()
        for key_name, config in self.mapping.items():
            parent_table = config.get("parent_table")
            if parent_table in self.metadata:
                continue
            path_to_metadata_storage = "model_artifacts/metadata"
            for file in os.listdir(path_to_metadata_storage):
                path_to_metadata_file = os.path.join(path_to_metadata_storage, file)
                metadata = MetadataLoader(path=path_to_metadata_file).load_data()
                if parent_table not in metadata:
                    continue
                # Validate the schema of the metadata of the parent table
                self._validate_schema()

                metadata.pop("global")
                self.merged_metadata.update(metadata)

                logger.info(
                    f"The metadata located in the path - '{path_to_metadata_storage}' "
                    f"has been merged with the current metadata as it contains "
                    f"the information of the parent table - '{parent_table}'"
                )
            self._check_merged_metadata(parent_table)

    def _check_existence_of_key_columns(self, table_name):
        """
        Check whether the key columns exist in the table
        """
        metadata_of_table = self.merged_metadata[table_name]
        existed_columns = self.existed_columns_mapping[table_name]
        for key, config_of_key in metadata_of_table.get("keys", {}).items():
            if all([column in existed_columns for column in config_of_key["columns"]]):
                continue
            else:
                non_existed_columns = [
                    f"{col!r}"
                    for col in set(config_of_key["columns"]).difference(set(existed_columns))
                ]
                message = (
                    f"The columns of the {config_of_key['type']} key '{key}' - "
                    f"{', '.join(non_existed_columns)} "
                    f"don't exist in the table - '{table_name}'"
                )
                self.errors["check existence of the key columns in 'columns'"][key] = message

    def _check_existence_of_referenced_columns(self, table_name: str):
        """
        Check whether the key columns exist in the referenced table
        """
        metadata_of_table = self.merged_metadata[table_name]
        for key, config_of_key in metadata_of_table.get("keys", {}).items():
            if config_of_key["type"] in self.type_of_fk_keys:
                referenced_table = config_of_key["references"]["table"]
                referenced_columns = config_of_key["references"]["columns"]
                existed_columns = self.existed_columns_mapping.get(referenced_table, {})
                non_existed_columns = [
                    f"{col!r}"
                    for col in referenced_columns
                    if col not in existed_columns
                ]
                if non_existed_columns:
                    message = (
                        f"The 'references.columns' of the {config_of_key['type']} '{key}' - "
                        f"{', '.join(non_existed_columns)} "
                        f"don't exist in the referenced table - '{referenced_table}'"
                    )
                    self.errors[
                        "check existence of the key columns in 'references.columns'"
                    ][key] = message

    def _fetch_existed_columns(self, table_name: str) -> List[str]:
        """
        Fetch the list of the columns of the source table
        """
        metadata_of_table = self.merged_metadata[table_name]
        format_settings = metadata_of_table.get("format", {})
        path_to_source = self._fetch_path_to_source(table_name)
        if path_to_source is not None:
            return self._get_columns_of_source(
                path_to_source=path_to_source,
                table_name=table_name,
                **format_settings
            )
        else:
            return self._get_columns_by_loader(table_name)

    def _gather_existed_columns(self, table_name: str):
        """
        Fetch the list of the existed columns of the table
        and put it into the mapping of existed columns
        """
        existed_columns = self._fetch_existed_columns(table_name)
        self.existed_columns_mapping[table_name] = existed_columns

    def _preprocess_metadata(self):
        """
        Preprocess the metadata, set the metadata and the merged metadata
        """
        self._define_mapping()
        self._merge_metadata()

    def _fetch_path_to_source(self, table_name: str):
        """
        Fetch the path to the source of the certain table
        """
        return self.merged_metadata[table_name].get("train_settings", {}).get("source")

    def _validate_fernet_key(self, table_name: str, fernet_key: str):
        """
        Validate the structure of the fernet key
        """
        try:
            DataEncryptor.validate_fernet_key(fernet_key)
        except ValueError as e:
            self.errors["validate structure of fernet key"][table_name] = str(e)

    def _check_access_to_input_data(self, table_name: str):
        """
        Check if the input data is accessible for the inference process
        """
        path_to_train_config = (
            f"model_artifacts/resources/{slugify(table_name)}/vae/checkpoints/train_config.pkl"
        )
        path_to_input_data = fetch_config(path_to_train_config).paths["input_data_path"]
        log_message = (
            "Access to a sample of the original data is required "
            "to generate reports during the inference process"
        )
        try:
            if self.loader is None:
                data_loader = DataLoader(
                    path=path_to_input_data,
                    table_name=table_name,
                    metadata=self.merged_metadata,
                    sensitive=True
                )
            else:
                data_loader = DataFrameFetcher(
                    loader=self.loader,
                    table_name=table_name
                )
            data_loader.get_columns()
        except InvalidToken:
            self.errors["check access to input data"][table_name] = (
                f"{log_message}. The provided Fernet key is invalid. "
                "The attempt to decrypt the sample of the original data has been failed. "
                "Please, provide the valid Fernet key."
            )
        except Exception as e:
            self.errors["check access to input data"][table_name] = f"{log_message}. {str(e)}"

    def _check_conditions_of_existence_of_input_data(self, table_name: str) -> bool:
        """
        Check if the input data is accessible for the inference process
        """
        return (
            not self.errors.get("validate structure of fernet key", {}).get(table_name)
            and not self.errors.get("check completion of the training process", {}).get(table_name)
        )

    def _launch_validation(self):
        """
        Launch the validation process
        """
        if self.type_of_process == "train":

            for table_name in self.merged_metadata.keys():
                self._gather_existed_columns(table_name)

            if self.errors:
                self._collect_errors()

        for table_name, table_metadata in self.metadata.items():
            fernet_key = table_metadata.get("encryption", {}).get("fernet_key")
            reports = table_metadata.get("infer_settings", {}).get("reports", [])

            if fernet_key is not None and (
                self.type_of_process == "train" or (
                    self.type_of_process == "infer" and reports
                )
            ):
                logger.warning(
                    f"Encryption and decryption are enabled for the table '{table_name}' "
                    "as a Fernet key is provided"
                )
                self._validate_fernet_key(table_name, fernet_key)

            if fernet_key is not None and self.type_of_process == "infer" and not reports:
                logger.warning(
                    f"The Fernet key is provided for the table '{table_name}' "
                    f"hasn't been validated because it will not be used"
                )

            if self.type_of_process == "train":
                self._check_existence_of_key_columns(table_name)
                self._check_existence_of_referenced_columns(table_name)

            elif self.type_of_process == "infer":
                self._check_completion_of_training(table_name)
                self._check_existence_of_destination(table_name)
                if reports and self._check_conditions_of_existence_of_input_data(table_name):
                    self._check_access_to_input_data(table_name)

        for table_name in self.metadata.keys():
            self._validate_metadata(table_name)

    def _collect_errors(self):
        """
        Collect the errors found during the validation process
        """
        error_logs = []
        for section, errors_details in self.errors.items():
            error_log = f'"{section}": {json.dumps(errors_details, indent=4)}'
            error_logs.append(error_log)
        if self.errors:
            message = (
                f"The validation of the metadata has been failed. "
                f"The error(s) found in - \n{', '.join(error_logs)}"
            )
            logger.error(message)
            raise ValidationError(message)
        logger.info("The validation of the metadata has been passed successfully")

    def run(self):
        """
        Run the validation process
        """
        self._preprocess_metadata()
        self._launch_validation()
        self._collect_errors()



from dataclasses import dataclass, field
from typing import Optional, Dict, List, Callable, Literal
from datetime import datetime

import pandas as pd
from slugify import slugify

from syngen.ml.utils import (
    slugify_attribute,
    fetch_unique_root,
    fetch_config,
    get_source_path_extension
)


@dataclass
class TrainConfig:
    """
    The configuration class to set up the work of train process
    """
    data: pd.DataFrame
    schema: Dict
    source: Optional[str]
    epochs: int
    drop_null: bool
    row_limit: Optional[int]
    table_name: str
    metadata: Dict
    metadata_path: Optional[str]
    reports: List[str]
    batch_size: int
    loader: Optional[Callable[[str], pd.DataFrame]]
    paths: Dict = field(init=False)
    row_subset: int = field(init=False)
    slugify_table_name: str = field(init=False)

    def __post_init__(self):
        self.row_subset = len(self.data)
        self.batch_size = min(self.batch_size, self.row_subset)
        self._set_paths()

    def __getstate__(self) -> Dict:
        """
        Return an updated config's instance
        """
        instance = self.__dict__.copy()
        attribute_keys_to_remove = ["data"]
        for attr_key in attribute_keys_to_remove:
            if attr_key in instance:
                del instance[attr_key]
        return instance

    def to_dict(self) -> Dict:
        """
        Return the values of the settings of training process
        """
        return {
            "epochs": self.epochs,
            "drop_null": self.drop_null,
            "row_subset": self.row_subset,
            "batch_size": self.batch_size,
            "reports": self.reports
        }

    @slugify_attribute(table_name="slugify_table_name")
    def _set_paths(self):
        """
        Create the paths which used in training process
        """
        timestamp = slugify(datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f"))
        fernet_key = self.metadata[self.table_name].get("encryption", {}).get("fernet_key")
        source_extension = get_source_path_extension(path=self.source)
        self.paths = {
            "model_artifacts_path": "model_artifacts/",
            "resources_path": f"model_artifacts/resources/{self.slugify_table_name}/",
            "tmp_store_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/",
            "source_path": self.source,
            "reports_path": f"model_artifacts/resources/{self.slugify_table_name}/reports",
            "path_to_sample_report": f"model_artifacts/resources/{self.slugify_table_name}/"
                                     f"reports/sample-report.html",
            "input_data_path": (
                f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                f"input_data_{self.slugify_table_name}."
                f"{'dat' if fernet_key is not None else 'pkl'}"
                if self.loader is None else None
            ),
            "state_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/checkpoints",
            "train_config_pickle_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                                        f"checkpoints/train_config.pkl",
            "dataset_pickle_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                                   f"checkpoints/model_dataset.pkl",
            "fk_kde_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                           f"checkpoints/stat_keys/",
            "original_schema_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"original_schema_{self.slugify_table_name}.pkl",
            "initial_order_of_columns_path": (
                f"model_artifacts/tmp_store/{self.slugify_table_name}"
                f"/initial_order_of_columns_{self.slugify_table_name}.pkl"
            ),
            "path_to_merged_infer": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"merged_infer_{self.slugify_table_name}{source_extension}",
            "no_ml_state_path":
                f"model_artifacts/resources/{self.slugify_table_name}/no_ml/checkpoints/",
            "path_to_flatten_metadata":
                f"model_artifacts/system_store/flatten_configs/"
                f"flatten_metadata_{fetch_unique_root(self.table_name, self.metadata_path)}.json",
            "losses_path": f"model_artifacts/system_store/losses/"
                           f"losses-{self.slugify_table_name}-{timestamp}.csv",
            "generated_reports": {}
        }


@dataclass
class InferConfig:
    """
    The configuration class to set up the work of the inference process
    """

    destination: Optional[str]
    metadata: Dict
    metadata_path: Optional[str]
    size: int
    table_name: str
    run_parallel: bool
    batch_size: Optional[int]
    random_seed: Optional[int]
    reports: List[str]
    both_keys: bool
    log_level: str
    loader: Optional[Callable[[str], pd.DataFrame]]
    type_of_process: Literal["train", "infer"]
    slugify_table_name: str = field(init=False)
    paths: Dict = field(init=False)

    def __post_init__(self):
        self.__set_paths()
        self._set_up_batch_size()

    def to_dict(self) -> Dict:
        """
        Return the values of the settings of inference process
        """
        return {
            "size": self.size,
            "run_parallel": self.run_parallel,
            "batch_size": self.batch_size,
            "random_seed": self.random_seed,
            "reports": self.reports,
        }

    def _set_up_batch_size(self):
        """
        Set up "batch_size" of inference process
        """
        self.batch_size = (
            min(self.batch_size, self.size) if self.batch_size is not None else self.size
        )

    @property
    def train_config(self):
        """
        Fetch the training configuration
        """
        return fetch_config(self.paths["train_config_pickle_path"])

    def _set_paths(self):
        """
        Create the paths which used in inference process
        """
        self.paths.update({
            "original_schema_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"original_schema_{self.slugify_table_name}.pkl",
            "path_to_flatten_metadata":
                f"model_artifacts/system_store/flatten_configs/"
                f"flatten_metadata_{fetch_unique_root(self.table_name, self.metadata_path)}.json",
            "input_data_path": self.train_config.paths["input_data_path"],
            "path_to_accuracy_report": (
                "model_artifacts/"
                f"{'tmp_store' if self.type_of_process == 'infer' else 'resources'}"
                f"/{self.slugify_table_name}/reports/accuracy-report.html"
            ),
            "generated_reports": {}
        })

    @slugify_attribute(table_name="slugify_table_name")
    def __set_paths(self):
        """
        Create the paths which used in inference process
        """
        dynamic_name = (
            self.slugify_table_name[:-3] if self.both_keys else self.slugify_table_name
        )
        self.paths = {
            "train_config_pickle_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/train_config.pkl"
        }
        source_extension = get_source_path_extension(
            path=self.train_config.paths["path_to_merged_infer"]
        )
        self.paths.update({
            "reports_path": (
                f"model_artifacts/"
                f"{'tmp_store' if self.type_of_process == 'infer' else 'resources'}"
                f"/{dynamic_name}/reports"
            ),
            "train_config_pickle_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/train_config.pkl",
            "infer_config_pickle_path":
                f"model_artifacts/tmp_store/{dynamic_name}/infer_config.pkl",
            "default_path_to_merged_infer": f"model_artifacts/tmp_store/{dynamic_name}/"
                                            f"merged_infer_{dynamic_name}{source_extension}",
            "path_to_merged_infer": (
                self.destination
                if self.destination is not None
                else f"model_artifacts/tmp_store/{dynamic_name}/"
                     f"merged_infer_{dynamic_name}{source_extension}"
            ),
            "state_path": f"model_artifacts/resources/{dynamic_name}/vae/checkpoints",
            "tmp_store_path": f"model_artifacts/tmp_store/{dynamic_name}",
            "vae_resources_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/",
            "dataset_pickle_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/model_dataset.pkl",
            "fk_kde_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/stat_keys/",
            "path_to_no_ml":
                f"model_artifacts/resources/{dynamic_name}/no_ml/checkpoints/kde_params.pkl"
        })

        self._set_paths()



from dataclasses import dataclass

import pandas as pd
from loguru import logger
from typing import Callable, Tuple, Dict, List


@dataclass
class DataFrameFetcher:
    """
    Class which fetches the dataframe by using the provided callback function
    """
    loader: Callable[[str], pd.DataFrame]
    table_name: str
    original_schema = None

    def load_data(self) -> Tuple[pd.DataFrame, Dict]:
        try:
            df = self.loader(self.table_name)
            default_schema = {"fields": {}, "format": "CSV"}
            logger.info(f"Successfully fetched dataframe for table: '{self.table_name}'")
            return df, default_schema
        except Exception as error:
            message = (
                "Failed to fetch the dataframe for the training process "
                "using the provided callback function. "
                f"Table: '{self.table_name}', Error: {error}"
            )
            logger.error(message)
            raise

    def get_columns(self, **kwargs) -> List[str]:
        """
        Get the column names of the table
        """
        df = self.loader(self.table_name)
        return df.columns.tolist()



from syngen.ml.utils.utils import (  # noqa: F401;
    get_date_columns,
    nan_labels_to_float,
    fillnan,
    slugify_attribute,
    slugify_parameters,
    inverse_dict,
    get_nan_labels,
    trim_string,
    check_if_features_assigned,
    convert_date_to_timestamp,
    convert_to_date,
    fetch_config,
    save_config,
    fetch_unique_root,
    generate_uuid,
    generate_uuids,
    setup_logger,
    datetime_to_timestamp,
    timestamp_to_datetime,
    fetch_timezone,
    get_log_path,
    setup_log_process,
    clean_up_metadata,
    fetch_log_message,
    file_sink,
    ProgressBarHandler,
    get_initial_table_name,
    get_reports,
    timing,
    ValidationError,
    TIMEZONE_REGEX,
    fetch_env_variables,
    get_source_path_extension
)



from typing import List, Dict
import base64
from collections import Counter
import re
import random

import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
from loguru import logger


def encode_categorical_features(dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
    """
    Encode categorical features in a list of DataFrames using Ordinal Encoder
    """
    encoder = OrdinalEncoder()
    dfs = [df.fillna("?") for df in dfs]

    encoder.fit(pd.concat(dfs, axis=0))
    res_dfs = []
    for df in dfs:
        encoded_df = encoder.transform(df)
        res_dfs.append(encoded_df)
    return res_dfs


def series_count_words(x):
    return len(str(x).split())


def set_empty_string(value):
    if value is None:
        return ""
    return str(value)


def text_to_continuous(df, text_columns: List[str]):
    """
    Add character length and word count columns to the DataFrame
    """
    for col in text_columns:
        df[col + "_char_len"] = df[col].apply(set_empty_string, convert_dtype=False).apply(len)
        df[col + "_word_count"] = (
            df[col].apply(set_empty_string, convert_dtype=False).apply(series_count_words)
        )
    return df


def transform_to_base64(path):
    """
    Encode the image to base64 format string
    """
    try:
        with open(path, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())
        return "data:image/svg+xml;base64," + encoded_string.decode("utf-8")
    except FileNotFoundError:
        logger.warning(f"No file found at {path}")
        return ""


def get_outlier_ratio_iqr(column: pd.Series, factor=1.5):
    """
    Get the ratio of outliers in a column using the IQR method
    """
    q1 = column.quantile(0.25)
    q3 = column.quantile(0.75)
    iqr = q3 - q1
    outlier_mask = (column < (q1 - factor * iqr)) | (column > (q3 + factor * iqr))
    return outlier_mask.mean()


def get_ratio_counts(ratio_counts, count: int = 30) -> Dict:
    """
    Select the most common, least common and some random items between them
    from the ratio_counts
    """
    ratio_counts = Counter(ratio_counts)

    most_least_count = int(count * 0.2)
    other_items = count - 2 * most_least_count

    most_common_items = ratio_counts.most_common(most_least_count)
    least_common_items = ratio_counts.most_common()[: -most_least_count - 1: -1]
    between_items = ratio_counts.most_common()[most_least_count:-most_least_count]
    selected_between_items = random.sample(between_items, min(other_items, len(between_items)))

    updated_ratio_counts = dict(
        most_common_items + selected_between_items + least_common_items
    )

    return updated_ratio_counts


def plot_dist(column_data, sort=True, full_set=None):
    """
    Plot the distribution of the column data
    """
    counts = Counter(column_data)
    if full_set is not None:
        absent_keys = full_set - set(counts.keys())
        if len(absent_keys) > 0:
            for k in absent_keys:
                counts[k] = 0

        if sort:
            sorted_keys = sorted(counts, key=lambda x: counts[x])
            counts = {i: counts[i] for i in sorted_keys}

        size = len(column_data)
        counts = {key: (x / size * 100) for key, x in counts.items()}
        return counts


def sanitize_labels(label):
    """
    Sanitize labels by removing dollar signs and carets
    """
    return re.sub(r"\$|\^", "", label)
